# Copyright (c) 2013 BaseSoft B.V.
# See LICENSE for details.

import datetime

from dateutil.relativedelta import relativedelta
from types import NoneType
from decimal import Decimal

from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.sql.expression import and_, func
import re

from django.utils.translation import ugettext as _
from django.core.exceptions import ValidationError
from django.conf import settings

from basesoft.util.virtual_date import today

from basesoft.web.core.util.controllers import transaction_decorator, LicenseConfigController
from basesoft.web.contrib.license.controllers import LicenseController
from basesoft.web.contrib.license.models import Domain, License
from basesoft.web.contrib.contact.models import Person, Company
from basesoft.web.core.exceptions import PublicUserException, SecurityException, PermissionDeniedException
from basesoft.web.contrib.i18n.controllers import I18nController
from basesoft.web.contrib.file.models import FileTmp
from basesoft.web.contrib.file.controllers import ProcessFileController, UploadFileController, AnonymousFileController
from basesoft.web.contrib.notification.controllers import NotificationController
from basesoft.web.contrib.i18n.models import EU_COUNTRIES, Country, EFTA_COUNTRIES
from basesoft.web.core.input.validation.plugins import Validation, Regex

from pvcycle.core.models import PvcTbsLicense, PvcAccountantTbs, PvcAccountantLicense, PvcTbsServiceLevel, \
    PvcMemberLicense, PvcTbsMember, PvcTbsMemberContractFile, \
    PvcTbsMemberServiceLevel, PvcOperatorLicense,\
    PvcTbsServiceLevelAdministrationFee, PvcTbsServiceLevelContributionFee,\
    PvcTbsCertificate, PvcTbsMemberCertificate, PvcTbsMembership, PvcTbsMembershipFee, PvcTbsMemberMembership, \
    PvcROECountry
from pvcycle.core.exceptions import PvcCertificateNotFoundException, PvcCertificateInvalidException
from pvcycle.subrogation.models import PvcSubrogation

class UniqueVATNumber(Validation):
    def check(self, params={"db_session": None}):

        if self._value in [None, '']:
            return

        if params['db_session'] is None:
            from basesoft.web.core.util.db import Session
            db_session = Session()
        else:
            db_session = params['db_session']

        _filter = Company.vat_no==self._value
        if params.has_key('license_id') and params['license_id'] != None:
            _filter = and_(_filter, PvcMemberLicense.id!=params['license_id'])

        a = db_session.query(License).with_polymorphic([PvcMemberLicense, PvcTbsLicense]).join(Company, License.company_id==Company.id).filter(_filter).first()

        if a is not None:
            raise ValidationError(_('A company is already registered with the provided VAT number.'))


class FrRCSLocationValidation(Regex):

    compiled_regex = re.compile(r"^[\w_\- ]+$", re.UNICODE)

    def check(self, params=None):
        self._handles['no_match'] = _('Value can only contain letters, underscores, dashes and spaces')

        if re.search('\d', unicode(self._value)):
            raise ValidationError(self._handles['no_match'])

        Regex.check(self, params)


class AbstractPvcController:

    def get_eu_countries_required_vat(self):
        # The UK DTS members are not obliged to have a VAT Number.
        eu_countries = [c for c in EU_COUNTRIES]
        eu_countries.remove('GB')
        return self.sa_session.query(Country).filter(Country.iso_3166_a2.in_(eu_countries)).all()

    def get_eu_countries(self):
        return self.sa_session.query(Country).filter(Country.iso_3166_a2.in_(EU_COUNTRIES)).all()

    def get_efta_countries(self):
        return self.sa_session.query(Country).filter(Country.iso_3166_a2.in_(EFTA_COUNTRIES)).all()

    def get_roe_countries(self):
        return [c.country for c in self.sa_session.query(PvcROECountry).all()]

    def _get_recipients_for_tbs_member(self, tbs_member):
        """
        Get the TBS groups and operator groups for the supplied tbs_member
        """
        operators = self.sa_session.query(PvcOperatorLicense).filter(PvcOperatorLicense.domain==tbs_member.tbs.domain).all()
        recipients = self.get_groups_in_license_with_permissions('pvc_tbs_member_manager', license_id=tbs_member.tbs_id)
        return recipients + [o.default_group for o in operators]

    def _get_take_back_schemes(self, id=None, permissions=None):
        """
        returns the tbs's the current user has access to
        """

        current_license = self._get_default_group().license

        # with 'pvc_tbs_license' return all tbs for the current domain (operator)
        if self.has_permissions('pvc_tbs_license'):

            q = self.sa_session.query(PvcTbsLicense).filter(PvcTbsLicense.domain == current_license.domain)

            if id is None:
                return q.all()

            return q.filter(PvcTbsLicense.id == id).all()

        tbs = []

        # get licenses with permission and check if they are of TBS type.
        if permissions is not None:
            licenses = self.get_account_licenses_with_permssions(*permissions)
            for license in licenses:
                if isinstance(license, PvcTbsLicense):
                    if id is None or license.id == id:
                        tbs.append(license)
            return tbs

        # check if this is a user who is part of TBS license(s)
        for ga in self.get_account().groups:
            if isinstance(ga.group.license, PvcTbsLicense):
                if id is None or ga.group.license.id == id:
                    tbs.append(ga.group.license)

        # if account is a member, check for TBS which member has applied to.
        if self.has_permissions('pvc_member_license'):

            member_licenses = self._get_member_licenses()
            for ml in member_licenses:
                tbs_members = self.sa_session.query(PvcTbsMember).filter(PvcTbsMember.member_id == ml.id)

                for tm in tbs_members:
                    if tm.tbs not in tbs:
                        tbs.append(tm.tbs)



        return tbs


    def _get_member_licenses(self, id=None, permissions=None):

        if id is not None:
            if type(id) != list:
                id = [id]

        if permissions is None:
            permissions = []

        # if we are operator (view_domain) get all PvcMemberLicenses within our domain
        if self.has_permissions('view_domain'):
            _filter = PvcMemberLicense.domain_id == self.get_default_group().license.domain.id

            if id is not None:
                _filter = and_(_filter, PvcMemberLicense.id.in_(id))

            return self.sa_session.query(PvcMemberLicense).filter(_filter).distinct().all()


        # else if we have pvc_tbs_member_manager permissions get all members for our TBS
        if self.has_permissions('pvc_tbs_member_manager'):
            _filter = PvcTbsMember.tbs_id.in_([t.id for t in self._get_take_back_schemes()])

            if id is not None:
                _filter = and_(_filter, PvcMemberLicense.id.in_(id))

            return self.sa_session.query(PvcMemberLicense).join(PvcTbsMember).filter(_filter).distinct().all()

        permissions = permissions + ['pvc_member_license']

        # else get our own MemberLicenses
        rv = self.get_account_licenses_with_permssions(*permissions)
        if id is None:
            return rv

        rv = [license for license in rv if license.id in id]
        if len(id) > len(rv):
            raise SecurityException('Trying to get Licenses without proper permissions %s, %s' % (id, permissions))
        return rv

    def is_commissioner(self):
        self.assert_permissions('pvc_member_license')
        for tbs_member in self._get_tbs_member_licenses():
            if tbs_member.is_commissioner:
                return True
        return False

    def has_active_subrogations(self, tbs_member=None):

        tbs_members = self._get_tbs_member_licenses()
        if tbs_member is not None:
            assert tbs_member in tbs_members
            tbs_members = [tbs_member]

        for tbs_member in tbs_members:
            for subrogation in tbs_member.subrogations:
                if subrogation.status == PvcSubrogation.STATUS_ACTIVE:
                    return True
        return False

    def _get_tbs_member_licenses(self, _id=None, permissions=None):
        members = [m.id for m in self._get_member_licenses(permissions=permissions)]
        if len(members) == 0:
            return []

        _filter = PvcTbsMember.member_id.in_(members)
        if _id is not None:
            _filter = and_(_filter, PvcTbsMember.id==_id)
        tbs_members = self.sa_session.query(PvcTbsMember).filter(_filter).all()
        if len(tbs_members) == 0:
            return []

        tbss = self._get_take_back_schemes()
        rv = []
        for tbs_member in tbs_members:
            if tbs_member.tbs not in tbss:
                if _id is not None:
                    raise SecurityException('Trying to get TBS Members without proper permissions %s' % _id)
                continue
            rv.append(tbs_member)

        return rv

    def _get_tbs_member_countries(self, tbs_member):

        countries = []

        efta = self.get_roe_countries()

        tbss = self.sa_session.query(PvcTbsLicense).filter(and_(PvcTbsLicense.is_active==True, PvcTbsLicense.domain==self.get_default_group().license.domain)).all()

        specific = [tbs.country for tbs in tbss if tbs.country is not None]
        leftover = [c for c in efta if c not in specific]

        tbs = tbs_member.tbs
        if tbs.country is None:
            countries += leftover
        else:
            countries.append(tbs.country)

        return countries

    def get_tbs_member_countries(self, tbs_member):
        return self._get_tbs_member_countries(tbs_member)

    def get_available_tbs_countries(self):
        efta = self.get_efta_countries()
        tbss = self.sa_session.query(PvcTbsLicense).filter(and_(PvcTbsLicense.is_active==True, PvcTbsLicense.domain==self.get_default_group().license.domain)).all()

        specific = [tbs.country for tbs in tbss if tbs.country is not None]
        return [c for c in efta if c not in specific]

class PvcOperatorController(LicenseController, AbstractPvcController):
    def get_domain_licenses(self, domain_id=None):
        return self._get_take_back_schemes()

class PvcNotificationController(NotificationController, AbstractPvcController):

    def has_recipient_permissions(self, recipient, for_license, for_domain):
        # Operator can send to anyone within the Domain.
        if isinstance(self.get_default_group().license, PvcOperatorLicense):
            if recipient.license.domain == self.get_default_group().license.domain:
                return recipient
        if isinstance(self.get_default_group().license, PvcTbsLicense):
            try:
                self.sa_session.query(PvcTbsMember).filter(and_(
                    PvcTbsMember.tbs == self.get_default_group().license,
                    PvcTbsMember.member == recipient.license,
                )).one()
            except NoResultFound:
                raise PermissionDeniedException('Insufficient permissions')
            return recipient

        # TODO: implement PvcMember Notification sending if required.

        raise PermissionDeniedException('Insufficient permissions')

class PvcTBSController(LicenseController, AbstractPvcController):

    def load_tbs(self, license_id):
        assert isinstance(license_id, (int, long))
        self.assert_permissions('pvc_tbs_license')

        state = self.load_license_state(license_id)

        license = self.get_licenses(license_id)[0]
        state['tbs'] = {
            'tbs_type': license.tbs_type,
            'contract_text': license.contract_text,
            'mandate_text': license.mandate_text,
            'terms_and_conditions_text': license.terms_and_conditions_text,
            'country': None
        }
        if license.country is not None:
            state['tbs']['country'] = license.country.id

        return state


    def get_take_back_schemes(self, id=None, permissions=None):
        """
        returns a list of tbs's for the current tbs license
        """

        return self._get_take_back_schemes(id=id, permissions=permissions)

    @transaction_decorator
    def save_tbs(self, state):
        assert isinstance(state, dict)
        self.assert_permissions('pvc_tbs_license')

        state['license'] = {
            'polymorphic_type': 'PvcTbsLicense',
            'name': state['company']['name'],
            'expiration_date': None,
            'is_default': 0
        }

        state['domain'] = {
            'name': self.get_default_group().license.domain.name
        }

        # TODO: make 2 groups for read/write permissions
        state['group'] = {
            'name': 'default'
        }
        state['permissions'] = []

        _license = self.save_license_from_state(state)

        if state.has_key('configs'):
            lcc = LicenseConfigController(self.sa_session, scope=_license)
            # if we use Memberships instead of Service Levels to determine the Administration Fee.
            if lcc.get('pvc.tbs.use.membership.instead.of.service.level'):
                self.grant_from_role('PVC DTS TBS', _license, _license.groups[0])
            else:
                self.grant_from_role('PVC TBS Admin', _license, _license.groups[0])

        return _license

    @transaction_decorator
    def save_roe_countries(self, values):
        self.assert_permissions("pvc_tbs_license")

        changed = False
        current = self.get_roe_countries()
        allowed = self.get_available_tbs_countries()
        newlist = self.sa_session.query(Country).filter(Country.id.in_(values)).all()
        for country in newlist:
            if country not in allowed:
                raise PublicUserException(_('This supplied country %(country)s is not allowed in Rest of Europe' % {'country': country.name}))
            if country not in current:
                changed = True
                self.logger.info('Adding country %(country)s to Rest of Europe list' % {'country': country.name})
                rc = PvcROECountry()
                rc.country = country
                self.sa_session.add(rc)

        for country in current:
            if country not in newlist:
                changed = True
                self.logger.info('Removing country %(country)s from Rest of Europe list' % {'country': country.name})
                self.sa_session.query(PvcROECountry).filter(PvcROECountry.country==country).delete()

        if changed:
            return True
        return False

    def refresh_roe_countries_for_pv_modules(self):
        """
        Call this after different controller methods affect the ROE countries, so it is outside of the transaction decorator.
        """
        import xmlrpclib
        from twisted.web.xmlrpc import Fault

        client = xmlrpclib.Server(settings.SCHEDULER_SERVER_XMLRPC, use_datetime=True, allow_none=True)
        try:
            client.start_task('pvcycle.pvmodule.schedule.DisablePVModuleTask', {})
        except Fault, e:
            raise PublicUserException(e.faultString)

    @transaction_decorator
    def save_tbs_texts(self, _id, state):

        self.assert_permissions("pvc_tbs_license")

        _license = self.sa_session.query(PvcTbsLicense).filter(PvcTbsLicense.id == _id).first()
        if _license is None:
            raise PublicUserException(_('TBS not found'))

        if state.has_key('contract_text'):
            _license.contract_text = state['contract_text']
        if state.has_key('mandate_text'):
            _license.mandate_text = state['mandate_text']
        if state.has_key('terms_and_conditions_text'):
            _license.terms_and_conditions_text = state['terms_and_conditions_text']
        self.sa_session.add(_license)
        self.sa_session.flush()
        return _license

    @transaction_decorator
    def save_or_update_license(self, id, name, domain, polymorphic_type, contact,
                               company, expiration_date, is_default=False, permissions=[], usage_options=[], state=None):

        assert isinstance(domain, Domain)
        assert isinstance(contact, Person)
        assert isinstance(company, Company)

        self.assert_permissions("pvc_tbs_license")

        if id is not None:

            # a license type cannot be changed; id and polymorphic type must match otherwise
            # next query will not have any results
            license = self.sa_session.query(polymorphic_type).filter(polymorphic_type.id == id)
            license = license.one()
        else:
            license = polymorphic_type()
            license.domain = domain

        assert isinstance(license, PvcTbsLicense)

        license.name = unicode(name)
        license.contact_person = contact
        license.expiration_date = expiration_date
        license.is_default = is_default
        license.company = company

        license.tbs_type = int(state['tbs']['tbs_type'])
        license.contract_text = None if not state['tbs'].has_key('contract_text') else state['tbs']['contract_text']

        if state['tbs']['country'] is not None:
            try:
                cid = int(state['tbs']['country'])
                # get the country
                i18n = I18nController(self.sa_session, account_id=self.get_account().id)
                license.country = i18n.get_countries(cid).pop()
                roe = self.sa_session.query(PvcROECountry).filter(PvcROECountry.country==license.country).first()
                if roe is not None:
                    self.logger.info('Deleting %(country)s from Rest of Europe' % {'country': roe.country.name})
                    self.sa_session.delete(roe)

            except ValueError:
                license.country = None
        else:
            license.country = None

        # And save the permissions to the license
        if len(permissions) > 0:
            self.save_permissions_to_license(license, permissions, None)

        self.sa_session.add(license)
        return license

    #
    # memberships
    #

    @transaction_decorator
    def save_membership_state(self, membership_id, state):

        self.logger.info('save_membership_state: %s' % (state))

        tbs = self.sa_session.query(PvcTbsLicense).filter(PvcTbsLicense.id == state['tbs_id']).one()
        name = state['name']

        assert isinstance(tbs, PvcTbsLicense)
        assert isinstance(name, (str, unicode))

        if not self.has_permissions('pvc_tbs_license'):
            self.assert_permissions("pvc_tbs_membership_level", license_id=tbs.id)
            pass

        # verify that this TBS has indeed no Service Levels
        lcc = LicenseConfigController(self.sa_session, scope=tbs)
        if not lcc.get('pvc.tbs.use.membership.instead.of.service.level'):
            raise PublicUserException(_('This TBS does not have membership fees enabled.'))

        m = self.sa_session.query(PvcTbsMembership).filter(
            and_(PvcTbsMembership.id!=membership_id, PvcTbsMembership.tbs==tbs, PvcTbsMembership.name==name)
        ).first()
        if m is not None:
            raise PublicUserException(_('A Membership with this name already exists at this TBS'))

        if membership_id is None:
            m = PvcTbsMembership()
            m.tbs = tbs
        else:
            m = self.sa_session.query(PvcTbsMembership).filter(PvcTbsMembership.id==membership_id).one()

        m.name = name
        self.sa_session.add(m)

        min_fee_year = today().year if m.id is None else today().year + 1

        # we can only add membership fees every 3 years as per spec.
        membership_years = lcc.get('pvc.tbs.membership.valid.for.years')

        previous_year = None
        for i in range(len(state['membership_fees'])):

            fee = state['membership_fees'][i]

            assert isinstance(fee['start_year'], (int, long))
            assert isinstance(fee['fee'], Decimal)

            try:
                m_fee = m.membership_fees[i]
            except IndexError:
                m_fee = PvcTbsMembershipFee()
            m_fee.membership = m

            # check if the user wants to change the fee
            if m_fee.start_year != fee['start_year'] or m_fee.fee != fee['fee']:

                # error if the user tries to change an existing fee from the past
                if m_fee.start_year is not None and m_fee.start_year <= today().year:
                    raise PvcInvalidFeeException(_('Cannot change fee in the past: %s') %  m_fee.start_year)

                # error if the user sets a date before the next declaration period
                if membership_id is not None and fee['start_year'] < min_fee_year:
                    raise PvcInvalidFeeException(_('Start year must be at least: %s') %  min_fee_year)

                # check that the new fee is minimally "membership_years" larger than previous year.
                # memberships are valid for X years.
                if previous_year is not None and fee['start_year'] < previous_year+membership_years:
                    raise PublicUserException(_('Membership fees have to be at least %s years apart.' % membership_years))

                previous_year = fee['start_year']

                m_fee.start_year = fee['start_year']
                m_fee.fee = fee['fee']

                self.logger.info('updated membership fee for membership=%s, fee=%s' % (m.id, m_fee.fee))
                self.sa_session.add(m_fee)

        # remove administration fee's that are not posted
        for i in range( len(state['membership_fees']), len(m.membership_fees) ):
            m_fee = m.membership_fees[i]

            if m_fee.start_year <= today().year:
                raise PvcInvalidFeeException(_('Cannot remove fee from the past: %s') %  m_fee.start_year)

            self.logger.info('removed membership fee for membership=%s, fee=%s' % (m.id, m_fee.fee))
            self.sa_session.delete(m_fee)

        return m

    #
    # service level
    #

    @transaction_decorator
    def save_service_level_state(self, service_level_id, state):

        self.logger.info('save_service_level_state: %s' % (state))

        tbs = self.sa_session.query(PvcTbsLicense).filter(PvcTbsLicense.id == state['tbs_id']).one()
        name = state['name']
        contract_annex_text = state['contract_annex_text']

        assert isinstance(tbs, PvcTbsLicense)
        assert isinstance(name, (str, unicode))

        # operator or own tbs
        if not self.has_permissions('pvc_tbs_license'):
            self.assert_permissions("pvc_tbs_service_level", license_id=tbs.id)

        #
        lcc = LicenseConfigController(self.sa_session, scope=tbs)
        if lcc.get('pvc.tbs.use.membership.instead.of.service.level'):
            raise PublicUserException(_('This TBS does not have service levels enabled.'))

        # check if unique name per tbs
        sls = self.sa_session.query(PvcTbsServiceLevel) \
            .filter(PvcTbsServiceLevel.id != service_level_id, PvcTbsServiceLevel.tbs_id == state['tbs_id'], PvcTbsServiceLevel.name == state['name']) \
            .all()
        if len(sls) > 0:
            raise PublicUserException(_('A Service Level with this name already exists for this TBS'))

        if service_level_id is None:
            sl = PvcTbsServiceLevel()
            sl.tbs = tbs
        else:
            sl = self.sa_session.query(PvcTbsServiceLevel) \
                .filter(PvcTbsServiceLevel.id == service_level_id) \
                .one()

        sl.name = name
        sl.contract_annex_text = contract_annex_text
        self.sa_session.add(sl)

        # allow any year on first save, else must be this year
        min_administration_fee_year = today().year if sl.id is None else today().year

        tbs_declaration_period = lcc.get('pvc.tbs.declaration.period')

        min_contribution_fee_date = None
        if tbs_declaration_period is None:
            raise PublicUserException(_('TBS must have a declaration period set'))
        elif tbs_declaration_period == PvcTbsLicense.DECLARATION_PERIOD_MONTHLY:

            next_month = today() + relativedelta(months=1)
            min_contribution_fee_date = datetime.date(next_month.year, next_month.month, 1)

        elif tbs_declaration_period == PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY:

            # get closest quarter
            for i in range(4):
                x = datetime.date(today().year, (i*3)+1, 1)
                if x > today():
                    min_contribution_fee_date = x
                    break

            # else take next year
            if not min_contribution_fee_date:
                min_contribution_fee_date = datetime.date(today().year+1, 1, 1)

        elif tbs_declaration_period == PvcTbsLicense.DECLARATION_PERIOD_YEARLY:
            min_contribution_fee_date = datetime.date(today().year+1, 1, 1)

        else:
            raise Exception('Unknown declaration period: %s' % tbs_declaration_period)


        #
        # insert/update administration fee
        #

        for i in range(len(state['administration_fees'])):

            fee = state['administration_fees'][i]

            assert isinstance(fee['start_year'], (int, long))
            assert isinstance(fee['fee'], Decimal)

            try:
                db_fee = sl.administration_fees[i]
            except IndexError:
                db_fee = PvcTbsServiceLevelAdministrationFee()
            db_fee.service_level = sl

            # check if the user wants to change the fee
            if db_fee.start_year != fee['start_year'] or db_fee.fee != fee['fee']:

                # error if the user tries to change an existing fee from the past
                if db_fee.start_year is not None and db_fee.start_year < today().year:
                    raise PvcInvalidFeeException(_('Cannot change fee in the past: %s') %  db_fee.start_year)

                # error if the user sets a date before the next declaration period
                if service_level_id is not None and fee['start_year'] < min_administration_fee_year:
                    raise PvcInvalidFeeException(_('Start year must be at least: %s') %  min_administration_fee_year)

                db_fee.start_year = fee['start_year']
                db_fee.fee = fee['fee']

                self.logger.info('updated administration fee for service_level=%s, fee=%s' % (sl.id, db_fee.fee))
                self.sa_session.add(db_fee)

        # remove administration fee's that are not posted
        for i in range( len(state['administration_fees']), len(sl.administration_fees) ):
            db_fee = sl.administration_fees[i]

            if db_fee.start_year <= today().year:
                raise PvcInvalidFeeException(_('Cannot remove fee from the past: %s') %  db_fee.start_year)

            self.logger.info('removed administration fee for service_level=%s, fee=%s' % (sl.id, db_fee.fee))
            self.sa_session.delete(db_fee)


        #
        # insert/update contribution fee
        #

        for i in range(len(state['contribution_fees'])):

            fee = state['contribution_fees'][i]

            assert isinstance(fee['start_date'], (datetime.date, datetime.datetime))
            assert isinstance(fee['fee'], Decimal)

            # js sends datetimes, convert to date
            if isinstance(fee['start_date'], datetime.datetime):
                fee['start_date'] = fee['start_date'].date()

            try:
                db_fee = sl.contribution_fees[i]
            except IndexError:
                db_fee = PvcTbsServiceLevelContributionFee()
            db_fee.service_level = sl

            # check if the user wants to change the fee
            if db_fee.start_date != fee['start_date'] or db_fee.fee != fee['fee']:

                # error if the user tries to change an existing fee from the past
                if db_fee.start_date is not None and db_fee.start_date <= today():
                    raise PvcInvalidFeeException(_('Cannot change fee in the past: %s') %  db_fee.start_date)

                # error if the user sets a date before the next declaration period
                if service_level_id is not None and fee['start_date'] < min_contribution_fee_date:
                    raise PvcInvalidFeeException(_('Start date must be at least next declaration period: %s') %  min_contribution_fee_date)

                # check validity of declaration period start
                if tbs_declaration_period == PvcTbsLicense.DECLARATION_PERIOD_MONTHLY:

                    if fee['start_date'].day != 1:
                        raise PvcInvalidFeeException(_('Fee must be set on 1st day of the month'))

                elif tbs_declaration_period == PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY:

                    if fee['start_date'].day != 1 or fee['start_date'].month not in [(i*3+1) for i in range(4)]:
                        raise PvcInvalidFeeException(_('Fee must be set on 1st day of the quarter'))

                elif tbs_declaration_period == PvcTbsLicense.DECLARATION_PERIOD_YEARLY:

                    if fee['start_date'].day != 1 and fee['start_date'].month != 1:
                        raise PvcInvalidFeeException(_('Fee must be set on 1st of January'))

                db_fee.start_date = fee['start_date']
                db_fee.fee = fee['fee']

                self.logger.info('updated contribution fee for service_level=%s, fee=%s' % (sl.id, db_fee.fee))
                self.sa_session.add(db_fee)

        # remove contribution fee's that are not posted
        for i in range( len(state['contribution_fees']), len(sl.contribution_fees) ):
            db_fee = sl.contribution_fees[i]

            if db_fee.start_date <= today():
                raise PvcInvalidFeeException(_('Cannot remove fee from the past: %s') %  db_fee.start_date)

            self.logger.info('removed contribution fee for service_level=%s, fee=%s' % (sl.id, db_fee.fee))
            self.sa_session.delete(db_fee)

        return sl

    @transaction_decorator
    def update_active_service_level_fees(self, service_level_id, contribution_fee, administration_fee):
        self.assert_permissions('pvc_tbs_license')
        sl = self.get_service_level(service_level_id)

        # get the start year of the current active administration fee
        af = self.sa_session.query(PvcTbsServiceLevelAdministrationFee).filter(and_(
            PvcTbsServiceLevelAdministrationFee.service_level_id==sl.id,
            PvcTbsServiceLevelAdministrationFee.start_year<=today().year,
        )).order_by(PvcTbsServiceLevelAdministrationFee.start_year.desc()).first()
        if af is None:
            raise PublicUserException(_('There is currently no active administration fee'))

        start_year = af.start_year

        # get the start date of the current active contribution fee
        cf = self.sa_session.query(PvcTbsServiceLevelContributionFee).filter(and_(
            PvcTbsServiceLevelContributionFee.service_level_id==sl.id,
            PvcTbsServiceLevelContributionFee.start_date<=today(),
        )).order_by(PvcTbsServiceLevelContributionFee.start_date.desc()).first()
        if cf is None:
            raise PublicUserException(_('There is currently no active administration fee'))

        start_date = cf.start_date

        a = self.update_service_level_administration_fee(sl.id, start_year, administration_fee)
        c = self.update_service_level_contribution_fee(sl.id, start_date, contribution_fee)
        return a, c

    @transaction_decorator
    def update_service_level_contribution_fee(self, service_level_id, start_date, new_fee):
        self.assert_permissions('pvc_tbs_license')
        assert isinstance(new_fee, Decimal)

        # get the service level contribution fee entry.
        sl = self.get_service_level(service_level_id)
        db_fee = self.sa_session.query(PvcTbsServiceLevelContributionFee).filter(and_(
            PvcTbsServiceLevelContributionFee.service_level==sl,
            PvcTbsServiceLevelContributionFee.start_date==start_date
        )).first()
        if db_fee is None:
            raise PublicUserException(_('No contribution fee entry found for the supplied start date.'))

        # check that there is no different current active fee, as all Declarations should already have been closed.
        newer = self.sa_session.query(PvcTbsServiceLevelContributionFee).filter(and_(
            PvcTbsServiceLevelContributionFee.service_level==sl,
            PvcTbsServiceLevelContributionFee.start_date>start_date,
            PvcTbsServiceLevelContributionFee.start_date<=today()
        )).first()
        if newer is not None:
            raise PublicUserException(_('There is already a newer contribution fee active.'))

        if new_fee == db_fee.fee:
            return None

        self.logger.info('altering contribution fee from %s for service_level=%s, new_fee=%s, start_date=%s' % (db_fee.fee, sl.id, new_fee, start_date))

        # update related Declarations!
        from pvcycle.order.controllers import PvcDeclarationOrderController
        dc = PvcDeclarationOrderController(self.sa_session, account_id=self.get_account().id, request=self.request)
        affected = dc.update_with_new_fee(sl, new_fee, db_fee.fee)

        db_fee.fee = new_fee
        self.sa_session.add(db_fee)

        return affected

    @transaction_decorator
    def update_service_level_administration_fee(self, service_level_id, start_year, new_fee):
        self.assert_permissions('pvc_tbs_license')
        assert isinstance(new_fee, Decimal)

        # get the service level administration fee entry.
        sl = self.get_service_level(service_level_id)
        db_fee = self.sa_session.query(PvcTbsServiceLevelAdministrationFee).filter(and_(
            PvcTbsServiceLevelAdministrationFee.service_level==sl,
            PvcTbsServiceLevelAdministrationFee.start_year==start_year
        )).first()
        if db_fee is None:
            raise PublicUserException(_('No administration fee entry found for the supplied start year.'))

        # check that there is no different current active fee, as all Declarations should already have been closed.
        newer = self.sa_session.query(PvcTbsServiceLevelAdministrationFee).filter(and_(
            PvcTbsServiceLevelAdministrationFee.service_level==sl,
            PvcTbsServiceLevelAdministrationFee.start_year>start_year,
            PvcTbsServiceLevelAdministrationFee.start_year<=today().year
        )).first()
        if newer is not None:
            raise PublicUserException(_('There is already a newer administration fee active.'))

        if new_fee == db_fee.fee:
            return None

        self.logger.info('altering administration fee from %s for service_level=%s, new_fee=%s, start_year=%s' % (db_fee.fee, sl.id, new_fee, start_year))

        db_fee.fee = new_fee
        self.sa_session.add(db_fee)

        return True

    def get_service_level(self, id):
        sl = self.sa_session.query(PvcTbsServiceLevel) \
            .filter(PvcTbsServiceLevel.id == id).one()

        if not self.has_permissions('pvc_tbs_license'):
            tbs_licenses = self.get_account_licenses_with_permssions('pvc_tbs_service_level')
            if not sl.tbs_id in [x.id for x in tbs_licenses]:
                raise SecurityException()

        return sl

    def get_service_level_state(self, _id):
        sl = self.get_service_level(_id)

        administration_fees = []
        for fee in sl.administration_fees:
            administration_fees.append({
                'start_year': fee.start_year,
                'fee': fee.fee,
            })

        contribution_fees = []
        for fee in sl.contribution_fees:
            contribution_fees.append({
                'start_date': fee.start_date,
                'fee': fee.fee,
            })

        return {
            'tbs_id': sl.tbs.id,
            'name': sl.name,
            'contract_annex_text': sl.contract_annex_text,
            'administration_fees': administration_fees,
            'contribution_fees': contribution_fees
        }

    def get_membership(self, _id):
        m = self.sa_session.query(PvcTbsMembership) \
            .filter(PvcTbsMembership.id == _id).one()

        if not self.has_permissions('pvc_tbs_license'):
            tbs_licenses = self.get_account_licenses_with_permssions('pvc_tbs_membership_level')
            if not m.tbs_id in [x.id for x in tbs_licenses]:
                raise SecurityException()

        return m

    def get_membership_state(self, _id):
        m = self.get_membership(_id)

        membership_fees = []
        for fee in m.membership_fees:
            membership_fees.append({
                'start_year': fee.start_year,
                'fee': fee.fee,
            })

        return {
            'tbs_id': m.tbs.id,
            'name': m.name,
            'membership_fees': membership_fees,
        }

    @transaction_decorator
    def delete_service_level(self, id):
        sl = self.get_service_level(id)

        if sl.get_current_administration_fee() is not None:
            raise PublicUserException(_('Cannot delete Service level: An administration fee is already active'))

        if sl.get_current_contribution_fee() is not None:
            raise PublicUserException(_('Cannot delete Service level: A contribution fee is already active'))

        self.sa_session.delete(sl)

    def get_contract_replacement_fields(self):
        return ContractReplacementFieldsContext().get_fields()


    def get_sl_replacement_fields(self):
        return ServiceLevelReplacementFieldsContext().get_fields()


    # certificate

    @transaction_decorator
    def save_tbs_certificate_state(self, tbs_certificate_id, state):

        self.logger.info('save_tbs_certificate_state: [%s]: %s' % (tbs_certificate_id, state))

        tbs = self.sa_session.query(PvcTbsLicense).filter(PvcTbsLicense.id == state['tbs_id']).one()
        text = state['text']
        year = state['year']

        assert isinstance(tbs, PvcTbsLicense)
        assert isinstance(year, (int))

        # operator or own tbs
        if not self.has_permissions('pvc_tbs_license'):
            self.assert_permissions("pvc_manage_certificates", license_id=tbs.id)

        # unique check
        certs = self.sa_session.query(PvcTbsCertificate) \
            .filter(PvcTbsCertificate.id != tbs_certificate_id, PvcTbsCertificate.tbs_id == state['tbs_id'], PvcTbsCertificate.year == state['year']) \
            .all()
        if len(certs) > 0:
            raise PublicUserException(_('A Certificate for this TBS and year already exists'))

        if tbs_certificate_id is None:
            cert = PvcTbsCertificate()
            cert.tbs = tbs
        else:
            cert = self.sa_session.query(PvcTbsCertificate) \
                .filter(PvcTbsCertificate.id == tbs_certificate_id) \
                .one()

        cert.year = year
        cert.text = text
        self.sa_session.add(cert)

        return cert


    def get_tbs_certificate(self, id_):
        cert = self.sa_session.query(PvcTbsCertificate) \
            .filter(PvcTbsCertificate.id == id_).one()

        if not self.has_permissions('pvc_tbs_license'):
            tbs_licenses = self.get_account_licenses_with_permssions('pvc_manage_certificates')
            if not cert.tbs_id in [x.id for x in tbs_licenses]:
                raise SecurityException()

        return cert

    def get_tbs_certificate_state(self, id_):
        cert = self.get_tbs_certificate(id_)

        return {
            'tbs_id': cert.tbs.id,
            'year': cert.year,
            'text': cert.text,
        }

    @transaction_decorator
    def delete_tbs_certificate(self, id_):
        cert = self.get_tbs_certificate(id_)
        self.sa_session.delete(cert)

    def get_certificate_replacement_fields(self):
        return CertificateReplacementFieldsContext().get_fields()



class AbstractReplacementFieldsContext(object):

    def _abstract_getattr(self, name):
        if name == 'country':
            return self._member.company.addresses[0].address.country
        if name == 'member':
            return self._member.company

        if name == 'member_contact':
            return self._member.company.contact_person

        if name == 'member_address':
            return self._member.company.addresses[0].address
        raise AttributeError

    def _member_fields(self, prefix='member', prefix_country=False, labels=None):
        country = '%s_country' % prefix if prefix_country else 'country'

        if labels is None:
            labels = {
                'member': _('Company'),
                'member_contact': _('Company contact'),
                'member_address': _('Company address')
            }

        return [{
            'key': prefix,
            'label': labels['member'], # Participant as they name it in the contract..
            'submenu': [{
                'key': 'name',
                'label': _('Name')
            }, {
                'key': 'vat_no',
                'label': _('VAT number')
            }, {
                'key': 'fr_rcs_number',
                'label': _('RCS number')
            }, {
                'key': 'fr_rcs_location',
                'label': _('RCS location')
            }, {
                'key': 'fr_company_type',
                'label': _('Company type')
            }]
        }, {
            'key': '%s_contact' % prefix,
            'label': labels['member_contact'],
            'submenu': [{
                'key': 'title',
                'label': _('Title')
            },{
                'key': 'person_name',
                'label': _('Full name')
            },{
                'key': 'position',
                'label': _('Position')
            },{
                'key': 'email',
                'label': _('Email')
            },{
                'key': 'phone',
                'label': _('Phone')
            },{
                'key': 'fax',
                'label': _('Fax')
            }]
        }, {
            'key': '%s_address' % prefix,
            'label': labels['member_address'],
            'submenu': [{
                'key': country,
                'label': _('Country'),
                'submenu': [{
                    'key': 'name',
                    'label': _('Name')
                }]
            },{
                'key': 'postcode',
                'label': _('Postal code')
            }, {
                'key': 'street',
                'label': _('Street (line 1)')
            }, {
                'key': 'street_2',
                'label': _('Street (line 2)')
            },{
                'key': 'city',
                'label': _('City')
            }]
        }]

class ServiceLevelReplacementFieldsContext(AbstractReplacementFieldsContext):
    def __init__(self, member=None, service_level=None):
        self._member = member
        self._service_level = service_level

    def __getattr__(self, name):

        if name == 'service_level':
            return self._service_level

        return self._abstract_getattr(name)

    def get_fields(self):
        rv = self._member_fields()

        rv += [{
            'key': 'service_level',
            'label': _('Service level'),
            'submenu': [{
                'key': 'name',
                'label': _('Name')
            },{
                'key': 'current_admin_fee',
                'label': _('Administration fee')
            },{
                'key': 'current_contrib_fee',
                'label': _('Contribution fee')
            }]
        }]

        return rv



class OtherContext(object):
    def __getattr__(self, name):
        if name == 'current_date':
            return datetime.datetime.now().date()

        raise Exception('invalid')

class ContractReplacementFieldsContext(AbstractReplacementFieldsContext):
    def __init__(self, member=None, tbs=None, service_level=None):
        self._member = member
        self._tbs = tbs
        self._service_level = service_level

    def __getattr__(self, name):

        if name == 'service_level':
            return self._service_level
        elif name == 'other':
            return OtherContext()

        return self._abstract_getattr(name)

    def get_fields(self):
        rv = self._member_fields()

        rv += [{
            'key': 'service_level',
            'label': _('Service level'),
            'submenu': [{
                'key': 'text',
                'label': _('Annex text')
            }]
        },

        {
            'key': 'other',
            'label': _('Other'),
            'submenu': [{
                'key': 'current_date',
                'label': _('Current date')
            }]
        }

        ]

        return rv


class CertificateReplacementFieldsContext(AbstractReplacementFieldsContext):

    def __init__(self, member=None, tbs=None, certificate=None):
        self._member = member
        self._tbs = tbs
        self._certificate = certificate

    def __getattr__(self, name):

        if name == 'certificate':
            return self._certificate

        return self._abstract_getattr(name)

    def get_fields(self):
        rv = self._member_fields()

        rv += [{
            'key': 'certificate',
            'label': _('Certificate'),
            'submenu': [{
                'key': 'issue_date',
                'label': _('Issue date')
            },{
                'key': 'hash',
                'label': _('Unique ID')
            }]
        }]

        return rv

class PvcAccountantController(LicenseController):

    def get_take_back_schemes(self, _id=None, permissions=None):
        """
        returns a list of tbs's for the current tbs license
        """

        rv = [at.tbs for at in self.get_default_group().license.accountant_tbs]
        if _id is not None:
            return [tbs for tbs in rv if tbs.id == _id]
        return rv

    def get_tbs_master_data(self):
        self.assert_permissions('pvc_view_tbs_master_data')
        return [self._load_master_data_state_from_license(t) for t in self.get_take_back_schemes()]

    def get_member_master_data(self, tbs_id):
        self.assert_permissions('pvc_view_member_master_data')
        tbs = self.get_take_back_schemes(tbs_id)
        assert len(tbs) == 1
        tbs = tbs.pop()

        # Get Active Members Only.
        result = self.sa_session.query(PvcMemberLicense).join(PvcTbsMember).filter(
            and_(
                PvcTbsMember.tbs==tbs,
                PvcTbsMember.status.in_([PvcTbsMember.STATUS_APPROVED, PvcTbsMember.STATUS_RESIGNING, PvcTbsMember.STATUS_RESIGNED]),
                PvcMemberLicense.is_active==True
            )
        ).all()

        return [self._load_master_data_state_from_license(m) for m in result]

    def get_licenses(self, id=None, polymorphic_type=None):
        if isinstance(self.get_account().default_group.license, PvcAccountantLicense):
            return self.get_take_back_schemes(id)

        self.assert_permissions('pvc_accountant_license')

        _filter = None

        if id is not None:
            if type(id) != list:
                id = [id]

            _filter = PvcAccountantLicense.id.in_(id)

        q = self.sa_session.query(PvcAccountantLicense)
        if _filter is not None:
            q = q.filter(_filter)
        return q.all()

    def load_accountant(self, license_id):
        assert isinstance(license_id, (int, long))
        self.assert_permissions('pvc_accountant_license')

        state = self.load_license_state(license_id)


        license = self.get_licenses(license_id, PvcAccountantLicense).pop()

        tbs_licenses = []
        for at in license.accountant_tbs:
            tbs_licenses.append({
                'tbs_id': at.tbs.id,
                'start_date': at.start_date,
                'end_date': at.end_date
            })

        state['tbs_licenses'] = tbs_licenses
        return state


    @transaction_decorator
    def save_accountant(self, state):
        assert isinstance(state, dict)
        self.assert_permissions('pvc_accountant_license')

        self.logger.info('save_accountant: %s' % (state))

        state['license'] = {
            'polymorphic_type': 'PvcAccountantLicense',
            'name': state['company']['name'],
            'expiration_date': None,
            'is_default': 0
        }

        state['domain'] = {
            'name': self.get_default_group().license.domain.name
        }

        # TODO: make 2 groups for read/write permissions
        state['group'] = {
            'name': 'default'
        }
        state['permissions'] = []

        license = self.save_license_from_state(state)
        self.grant_from_role('PVC Accountant', license, group=license.default_group)

        # the supplied TBS should be linked. They can have a start date and an end date.
        if state.has_key('tbs_licenses') and self.has_permissions('pvc_tbs_license'):

            tbsc = PvcTBSController(self.sa_session, account_id=self.get_account().id)
            # does the link already exists
            keep = []
            for tbs_state in state['tbs_licenses']:
                schemes = tbsc.get_take_back_schemes(id=tbs_state['tbs_id'])
                if len(schemes) == 0:
                    continue
                tbs = schemes.pop()
                keep.append(tbs)
                try:
                    at = self.sa_session.query(PvcAccountantTbs).filter(and_(PvcAccountantTbs.tbs == tbs, PvcAccountantTbs.accountant == license)).one()
                except NoResultFound:
                    at = PvcAccountantTbs()
                    at.tbs = tbs
                    at.accountant = license

                if tbs_state['start_date'] == '':
                    tbs_state['start_date'] = None

                if tbs_state['end_date'] == '':
                    tbs_state['end_date'] = None

                at.start_date = tbs_state['start_date']
                at.end_date = tbs_state['end_date']
                self.sa_session.add(at)

            for tbs in license.schemes:
                # delete TBS from this license who no longer are linked to this accountant.
                if tbs not in keep:
                    self.sa_session.query(PvcAccountantTbs).filter(and_(PvcAccountantTbs.tbs == tbs, PvcAccountantTbs.accountant == license)).delete()

            self.sa_session.flush()
            self.sa_session.expire(license)

        return license

    @transaction_decorator
    def save_or_update_license(self, id, name, domain, polymorphic_type, contact,
                               company, expiration_date, is_default=False, permissions=[], usage_options=[], state=None):
        """
        Override method in order to skip original permission checks.
        """
        assert isinstance(domain, Domain)
        assert isinstance(contact, Person)
        assert isinstance(company, Company)

        self.assert_permissions("pvc_accountant_license")

        if id is not None:
            license = self.sa_session.query(polymorphic_type).filter(polymorphic_type.id == id)
            license = license.one()
        else:
            license = polymorphic_type()
            license.domain = domain

        assert isinstance(license, PvcAccountantLicense)

        license.name = unicode(name)
        license.contact_person = contact
        license.expiration_date = expiration_date
        license.is_default = is_default
        license.company = company

        # And save the permissions to the license
        if len(permissions) > 0:
            self.save_permissions_to_license(license, permissions, None)

        self.sa_session.add(license)
        return license

class PvcMemberController(LicenseController, AbstractPvcController):

    def has_filled_company_data(self, group=None):
        """
        should this be moved to iqportal?
        """

        if group is None:
            group = self.get_default_group()

        assert group is not None
        return group.license.company is not None

    def get_licenses(self, license_id=None, polymorphic_type=None, permissions=None):
        return self._get_member_licenses(license_id, permissions)

    def load_member_license(self, license_id):

        assert isinstance(license_id, (int, long))
        self.assert_permissions('pvc_member_license', license_id=license_id)

        state = self.load_license_state(license_id)

        return state

    @transaction_decorator
    def save_member_license(self, state):

        assert isinstance(state, dict)

        self.logger.info('save_member_license: %s' % (state))

        license_id = None if not state.has_key('dbID') else state['dbID']
        self.logger.info('license_id: %s' % license_id)

        if license_id is None:
            self.assert_permissions('pvc_member_license')
            license = self._create_new_license(name=state['company']['name'], polymorphic_type=PvcMemberLicense, role_name='PVC Simple Member')
        else:

            # check own license
            self.assert_permissions('pvc_member_license', license_id=state['dbID'])

            license = self.sa_session.query(License).filter(License.id == license_id).one()

            # TODO: assert not active at any tbs

            # check if its a temp license, next time use the master data change controller
            #group = self._get_default_group()
            if license.polymorphic_type == 'License':

                self.logger.info('converting license')

                license.polymorphic_type = 'PvcMemberLicense'
                license.name = state['company']['name']
                self.sa_session.add(license)
                self.sa_session.execute("insert into tblPvcMemberLicense (license_id) values (:license_id)", {'license_id': license_id})
                self.logger.info('License converted to PvcMemberLicense for license_id: %s' % license_id)

        # return new license
        self.logger.info('calling update_mster_data')
        license = self._update_master_data(license, state, auto_approve=True)

        # expunge the license so we dont get issues with PvcMemberLicense being a License later on
        self.sa_session.expunge(license)

        self.logger.info('returning license')

        return license

    def get_take_back_schemes(self, id=None, _type=None):
        """
        returns a list of tbs's for which the current (new) member can apply
        """

        self.assert_permissions('pvc_member_license')

        _filter = PvcTbsLicense.domain==self.get_default_group().license.domain

        if _type is not None:
            assert _type in [PvcTbsLicense.TBS_TYPE_LOCAL, PvcTbsLicense.TBS_TYPE_SUBSIDIARY]
            _filter = and_(_filter, PvcTbsLicense.tbs_type==_type)

        q = self.sa_session.query(PvcTbsLicense)
        if id is not None:
            _filter = and_(_filter, PvcTbsLicense.id == id)
            return q.filter(_filter).order_by(PvcTbsLicense.name).one()

        return q.filter(_filter).order_by(PvcTbsLicense.name).all()

    def get_active_for_tbs(self, member_id, is_producer=None, is_commissioner=None):

        ml = self._get_member_licenses(member_id)

        _filter = and_(PvcTbsMember.member_id.in_([m.id for m in ml]), PvcTbsMember.status==PvcTbsMember.STATUS_APPROVED)

        if is_producer is not None:
            _filter = and_(_filter, PvcTbsMember.is_producer==is_producer)

        if is_commissioner is not None:
            _filter = and_(_filter, PvcTbsMember.is_commissioner==is_commissioner)

        tbs_members = self.sa_session.query(PvcTbsMember).filter(_filter).all()
        return tbs_members

    def get_member_countries(self, _id=None, mapped=False, tbs_member_id=None):
        """
        returns all the countries linked from the TBS for the supplied member id.
        """
        assert isinstance(self.get_default_group().license, PvcMemberLicense)
        if _id is None:
            _id = self.get_default_group().license.id

        roe = self.get_roe_countries()
        specific = [tbs.country for tbs in self.get_take_back_schemes() if tbs.country is not None and tbs.is_active]
        leftover = [c for c in roe if c not in specific]

        _map = {}
        countries = []
        tms = None
        if tbs_member_id is None:
            tms = self.get_tbs_members(_id)
        else:
            tms = [self._get_tbs_member(tbs_member_id)]

        for tm in tms:
            tbs = tm.tbs
            if tbs.country is None:
                for country in leftover:
                    _map[country.id] = (tm, country, )
                countries += leftover
            else:
                _map[tbs.country.id] = (tm, tbs.country, )
                countries.append(tbs.country)

        if mapped:
            return _map

        return countries

    def get_memberships(self, tbs):
        """
        returns all service levels for the given tbs_id
        """

        self.assert_permissions('pvc_member_license')

        if isinstance(tbs, (int, long)):
            tbs = self.sa_session.query(PvcTbsLicense).filter(PvcTbsLicense.id == tbs).one()
        assert isinstance(tbs, PvcTbsLicense)

        return self.sa_session.query(PvcTbsMembership) \
            .filter(PvcTbsMembership.tbs == tbs) \
            .order_by(PvcTbsMembership.name).all()

    def get_service_levels(self, tbs):
        """
        returns all service levels for the given tbs_id
        """

        self.assert_permissions('pvc_member_license')

        if isinstance(tbs, (int, long)):
            tbs = self.sa_session.query(PvcTbsLicense).filter(PvcTbsLicense.id == tbs).one()
        assert isinstance(tbs, PvcTbsLicense)

        return self.sa_session.query(PvcTbsServiceLevel) \
            .filter(PvcTbsServiceLevel.tbs == tbs) \
            .order_by(PvcTbsServiceLevel.name).all()

    def get_service_level(self, service_level):

        self.assert_permissions('pvc_member_license')

        if isinstance(service_level, (int, long)):
            service_level = self.sa_session.query(PvcTbsServiceLevel).filter(PvcTbsServiceLevel.id == service_level).one()
        assert isinstance(service_level, PvcTbsServiceLevel)

        return service_level

    @transaction_decorator
    def apply_at_tbs(self, tbs_license_id, member_license_id, tbs_member_id, service_level_id, is_commissioner,
                     is_producer, upload_contract, was_member_in_2013=False, membership_id=None, company_reg_no=None, accepted_tac=None):
        """
        method called from the MemberDetailsModule by the member while applying from the wizard or uploading document later

        @param membership_id: should be supplied at TBS which has memberships instead of service levels
        @param company_reg_no: the Company Registration Number, sometimes required at a TBS
        @param accepted_tac: boolean if the user accepted the Terms and Conditions.
        """

        assert isinstance(tbs_license_id, (int, long))
        assert isinstance(member_license_id, (int, long))
        assert isinstance(service_level_id, (int, long, NoneType))
        assert isinstance(tbs_member_id, (int, long, NoneType))

        assert isinstance(is_commissioner, (bool))
        assert isinstance(is_producer, (bool))

        assert isinstance(was_member_in_2013, (bool))
        assert isinstance(upload_contract, (FileTmp, NoneType))

        # check current account has permissions for given member
        self.assert_permissions("pvc_member_license", license_id=member_license_id)

        member_license = self.sa_session.query(PvcMemberLicense).filter(PvcMemberLicense.id == member_license_id).first()
        if member_license is None:
            raise PublicUserException(_('You have to finish the Registration process before applying at a TBS.'))

        tbs_license = self.get_take_back_schemes(tbs_license_id)
        if tbs_license.tbs_type != PvcTbsLicense.TBS_TYPE_SUBSIDIARY:
            raise PublicUserException(_('Cannot apply at a local TBS.'))

        lcc = LicenseConfigController(self.sa_session, scope=tbs_license)

        # Check if the company country is UK and if the TBS allows empty vats
        if member_license.company.vat_no is None and member_license.company.addresses[0].address.country.iso_3166_a2 == 'GB' and not lcc.get('pvc.tbs.optional.uk.vat.no'):
            raise PublicUserException(_('This TBS requires a valid VAT Number for a UK company.'))

        # Check if the TBS requires the member to be in the same country.
        if lcc.get('pvc.tbs.require.member.in.own.country') and member_license.company.addresses[0].address.country != tbs_license.company.addresses[0].address.country:
            raise PublicUserException(_('This TBS only allows companies from within %(country_name)s.' % {'country_name': tbs_license.company.addresses[0].address.country.name}))

        # Check if the TBS requires the member to accept the terms and conditions.
        if lcc.get('pvc.tbs.require.accepting.of.terms.and.conditions') and not accepted_tac:
            raise PublicUserException(_('You have to accept the Terms and Conditions in order to continue.'))

        # Boolean if the TBS uses Membership instead of Service Levels to determine the Fee. #415
        is_dts_member = lcc.get('pvc.tbs.use.membership.instead.of.service.level')

        # at least one checkbox should be true
        if not is_dts_member:
            assert is_producer is True or is_commissioner is True
        else:
            assert isinstance(membership_id, (int, long))

        service_level = None
        membership = None
        if is_dts_member:
            membership = self.sa_session.query(PvcTbsMembership).filter(PvcTbsMembership.id == membership_id).one()
        elif is_producer:
            service_level = self.sa_session.query(PvcTbsServiceLevel).filter(PvcTbsServiceLevel.id == service_level_id).one()
            # check that the service level is from this tbs
            assert service_level.tbs == tbs_license

        # link member with tbs
        if tbs_member_id is None:

            # check if already applied
            if len(self.sa_session.query(PvcTbsMember).filter(PvcTbsMember.tbs == tbs_license, PvcTbsMember.member == member_license).all()) > 0:
                raise PublicUserException(_('The member is already applied at this TBS, you will be notified when your application have been approved inside the PV CYCLE extranet.'))

            tbs_member = PvcTbsMember()
            tbs_member.tbs = tbs_license
            tbs_member.member = member_license
        else:

            # user is updating existing PvcTbsMember
            tbs_member = self.sa_session.query(PvcTbsMember).filter(PvcTbsMember.id == tbs_member_id).one()
            assert tbs_member.tbs == tbs_license
            assert tbs_member.member == member_license

            # check if member is allowed to modify
            if not tbs_member.status in (PvcTbsMember.STATUS_INCOMPLETE, PvcTbsMember.STATUS_REJECTED):
                raise PublicUserException(_('Cannot modifify registration: status not incomplete or rejected'))

            # remove all service_levels, we'll add a new one below
            for x in self.sa_session.query(PvcTbsMemberServiceLevel).filter(PvcTbsMemberServiceLevel.tbs_member == tbs_member):
                self.sa_session.delete(x)


        if not lcc.get('pvc.tbs.was.active.in.2013'): # if the TBS was not active in 2013 we force False.
            was_member_in_2013 = False

        lc = LicenseController(self.sa_session, account_id=self.get_account().id, request=self.request)
        if not is_dts_member:
            lc.grant_from_role('PVC Member Admin', member_license, member_license.default_group, add_dashboard=True)
        else:
            lc.grant_from_role('PVC DTS Member', member_license, member_license.default_group, add_dashboard=True)

        if lcc.get('pvc.tbs.require.company.registration.number'):
            assert company_reg_no is not None
            tbs_member.company_registration_number = company_reg_no

        tbs_member.is_commissioner = is_commissioner
        tbs_member.is_producer = is_producer
        tbs_member.was_member_in_2013 = was_member_in_2013

        if is_producer:
            # if they were a member in 2013 the contract can be uploaded by Operator/TBS (#174)
            if upload_contract is None and was_member_in_2013 is False:
                tbs_member.status = PvcTbsMember.STATUS_INCOMPLETE
            else:
                tbs_member.status = PvcTbsMember.STATUS_PENDING

        else:
            # else must be a commissioner, set status pending
            tbs_member.status = PvcTbsMember.STATUS_PENDING

        if lcc.get('pvc.tbs.auto.approve.member'):
            tbs_member.status = PvcTbsMember.STATUS_APPROVED

        self.sa_session.add(tbs_member)

        # Only add a service level if we have one selected (#366)
        if service_level is not None:
            self._add_service_level_or_membership(PvcTbsMemberServiceLevel, tbs_member, service_level)
        elif membership is not None:
            self._add_service_level_or_membership(PvcTbsMemberMembership, tbs_member, membership)

        self.sa_session.flush()

        if upload_contract is not None:
            # save contract file using file controller
            file_controller = ProcessFileController(self.sa_session, account_id=self.account_id)
            fc = file_controller.get_category_by_label('tbs contracts')
            if fc is None:
                fc = file_controller.save_or_update_file_category(None, u'tbs contracts')
            self.sa_session.flush()
            extra = {
                'tbs_member': tbs_member,
                'type': PvcTbsMemberContractFile.TYPE_CONTRACT_FULL
            }
            iq_file = file_controller.save_files_from_temp([upload_contract], fc, polymorphic_type=PvcTbsMemberContractFile, extra=extra)[0]
            self.sa_session.add(iq_file)

        self.logger.info('Mermber license %s applied at tbs %s for service_level %s. producer=%s, is_commissioner=%s' % (member_license_id, tbs_license.id, service_level_id, is_producer, is_commissioner))

        # SEND A NOTIFICATION TO THE TBS AND OPERATORS. (#315)
        if tbs_member.status == PvcTbsMember.STATUS_PENDING:
            nc = PvcNotificationController(self.sa_session, account_id=self.get_account().id)
            operators = self.sa_session.query(PvcOperatorLicense).filter(PvcOperatorLicense.domain==tbs_member.tbs.domain).all()
            recipients = self.get_groups_in_license_with_permissions('pvc_tbs_member_manager', license_id=tbs_member.tbs_id)
            recipients = recipients + [o.default_group for o in operators]
            message = 'Member "%(member)s" has applied at TBS "%(tbs)s" and requires approval.' % {'member': tbs_member.member.name, 'tbs': tbs_member.tbs.name}
            notification = nc.set_notification('Member "%(member)s" has applied at "%(tbs)s"' % {'member': tbs_member.member.name, 'tbs': tbs_member.tbs.name}, message)
            nc.create_system_notification(notification, recipients, for_license=True)
        elif tbs_member.status == PvcTbsMember.STATUS_APPROVED:
            # trigger administrative stuff
            from pvcycle.order.controllers import PvcOrderController, PvcAdminFeeAlreadyExists

            oc = PvcOrderController(self.sa_session, account_id=self.get_account().id)
            try:
                order = oc.trigger_administration_fee_order(tbs_member)
            except PvcAdminFeeAlreadyExists:
                self.logger.info('Not creating administration fee, order already exists for %s' % today().year)

        return tbs_member

    def _add_service_level_or_membership(self, obj, tbs_member, backref):
        """
        Membership and Service Level have same behavior.
        """

        # if we already had a service level/membership set for this date we update it.
        try:
            _instance = self.sa_session.query(obj).filter(
                and_(obj.tbs_member==tbs_member, obj.start_date==today())).one()
        except NoResultFound:
            _instance = obj()
        _instance.tbs_member = tbs_member
        setattr(_instance, obj.BACKREF, backref)
        _instance.start_date = today()
        self.sa_session.add(_instance)
        return _instance

    def get_tbs_members(self, member_id):
        return self.sa_session.query(PvcTbsMember).filter(PvcTbsMember.member_id == member_id).all()

    def is_applied_at_tbs(self, member_license_id):

        assert isinstance(member_license_id, (int, long))
        self.assert_permissions("pvc_member_license", license_id=member_license_id)

        return self.sa_session.query(PvcTbsMember).filter(PvcTbsMember.member_id == member_license_id).count() > 0

    def get_unfinished_tbs_members(self, member_license_id):
        """
        returns true if the given member_id should show a wizard. this is called after the user logins or refreshes
        """
        assert isinstance(member_license_id, (int, long))
        self.assert_permissions("pvc_member_license", license_id=member_license_id)


        return self.sa_session.query(PvcTbsMember) \
            .filter(PvcTbsMember.member_id == member_license_id,
                    PvcTbsMember.status == PvcTbsMember.STATUS_INCOMPLETE
            ).all()

    def load_master_data_state(self, _id, original=False):

        if isinstance(_id, PvcMemberLicense):
            _id = _id.id

        _license = self._get_member_licenses(_id).pop()
        state = super(PvcMemberController, self).load_master_data_state(_license, original=original)

        state['company_js_module'] = 'pvc_company_form'
        state['person_js_module'] = 'pvc_person_form'
        state['address_js_module'] = 'pvc_address_form'
        state['bank_account_js_module'] = 'pvc_bank_account_form'
        if state.has_key('bank_data') and state['bank_data'].has_key('address_data'):
            state['company_bank_account_address'] = state['bank_data']['address_data']

        return state

    def update_master_data(self, id, state, auto_approve=None):

        # auto approve is exlpicit set in save_member_license()
        if auto_approve is None:

            auto_approve = False

            # if operator or tbs, approve if they have the license_masterdata_approval
            if self.has_any_permissions('pvc_tbs_license', 'pvc_operator_license'):
                auto_approve = self.has_permissions('license_masterdata_change_approval')

        license = self._get_member_licenses(id, permissions=['update_license_masterdata', ]).pop()

        if not self.has_permissions('license_masterdata_change_approval'):
            self.assert_permissions('update_license_masterdata', license_id=id)

        return super(PvcMemberController, self)._update_master_data(license, state, auto_approve=auto_approve)

    def _get_tbs_member(self, tbs_member_id):
        """
        Returns the PvcTbsMember if accesable
        """

        tbs_member = self.sa_session.query(PvcTbsMember).filter(PvcTbsMember.id == tbs_member_id).one()

        # retrieve the member to see if we have access to it
        self._get_member_licenses(tbs_member.member_id).pop()

        # same for tbs
        self._get_take_back_schemes(tbs_member.tbs_id).pop()

        return tbs_member

    def get_tbs_member(self, tbs_member_id):
        return self._get_tbs_member(tbs_member_id)

    def load_tbs_member(self, tbs_member_id):

        tbs_member = self._get_tbs_member(tbs_member_id)

        contract_files = self.sa_session.query(PvcTbsMemberContractFile) \
            .filter(PvcTbsMemberContractFile.tbs_member_id == tbs_member_id) \
            .order_by(PvcTbsMemberContractFile.created.desc()) \
            .all()

        contract_file_hash = None
        if len(contract_files):
            contract_file_hash = contract_files[0].file_hash

        lcc = LicenseConfigController(self.sa_session, scope=tbs_member.tbs)

        return {
            'tbs_id': tbs_member.tbs.id,
            'tbs_name': tbs_member.tbs.name,
            'member_id': tbs_member.member.id,
            'member_name': tbs_member.member.name,
            'service_levels': [{'id': x.service_level.id, 'name': x.service_level.name, 'start_date': x.start_date} for x in tbs_member.service_levels ],

            'is_producer': tbs_member.is_producer,
            'is_commissioner': tbs_member.is_commissioner,
            'is_representative': tbs_member.is_representative,
            'company_registration_number': tbs_member.company_registration_number,

            'status': tbs_member.status,
            'upload_contract': contract_file_hash,
            'view_contract': contract_file_hash,
            'is_full_member': tbs_member.is_full_member,
            'has_voting_rights': tbs_member.has_voting_rights,
            'number_of_votes': tbs_member.number_of_votes,
            'has_full_memberships': lcc.get('pvc.tbs.has.full.memberships'),
            'has_active_subrogations': self.has_active_subrogations(tbs_member),
        }

    def get_contract(self, tbs_member_id):
        # this checks rights
        tbs_member = self._get_tbs_member(tbs_member_id)

        contract_file = self.sa_session.query(PvcTbsMemberContractFile) \
            .filter(PvcTbsMemberContractFile.tbs_member_id == tbs_member_id) \
            .order_by(PvcTbsMemberContractFile.created.desc()) \
            .first()

        if contract_file is None:
            raise PublicUserException(_('No contract uploaded'))

        return contract_file.file_hash

    @transaction_decorator
    def save_tbs_member(self, tbs_member_id, state):
        """
        method used by 'member_details' interface to update the status from the tbs_member
        """

        tbs_member = self._get_tbs_member(tbs_member_id)

        old_status = tbs_member.status
        new_status = int(state['status'])

        if old_status != new_status:

            # TODO: check for a specific status update permission?

            #
            # switch current status
            #

            if old_status == PvcTbsMember.STATUS_INCOMPLETE:

                # TODO: check if member
                raise PublicUserException(_('Cannot change status: current status is Incomplete'))

            elif old_status == PvcTbsMember.STATUS_PENDING:

                if new_status not in (PvcTbsMember.STATUS_APPROVED, PvcTbsMember.STATUS_REJECTED):
                    raise PublicUserException(_('New status must be Approved or Rejected'))

            elif old_status == PvcTbsMember.STATUS_APPROVED:

                if new_status not in (PvcTbsMember.STATUS_SUSPENDED, PvcTbsMember.STATUS_RESIGNING,):
                    raise PublicUserException(_('New status can only be Suspended'))

            elif old_status == PvcTbsMember.STATUS_REJECTED:

                # TODO: only allow member to change to pending
                if new_status not in (PvcTbsMember.STATUS_PENDING, ):
                    raise PublicUserException(_('New status must be Pending'))

            elif old_status == PvcTbsMember.STATUS_SUSPENDED:

                if new_status not in (PvcTbsMember.STATUS_APPROVED,):
                    raise PublicUserException(_('New status must be Approved'))

            elif old_status == PvcTbsMember.STATUS_RESIGNING:

                if new_status not in (PvcTbsMember.STATUS_APPROVED,):
                    raise PublicUserException(_('New status must be Approved'))

            elif old_status == PvcTbsMember.STATUS_RESIGNED:

                if new_status not in (PvcTbsMember.STATUS_APPROVED,):
                    raise PublicUserException(_('New status must be Approved'))

            else:
                raise PublicUserException(_('Invalid status'))

            # prepare notification sending
            notification = None
            nc = PvcNotificationController(self.sa_session, account_id=self.get_account().id)
            recipients = self.get_groups_in_license_with_permissions('pvc_member_license', license_id=tbs_member.member_id)

            #
            # switch new status
            #

            if new_status == PvcTbsMember.STATUS_APPROVED:

                _message = _('You are now successfully registered as new member at %(tbs_name)s.' % {'tbs_name': tbs_member.tbs.name})
                _message = '%s\n\n%s' % (_message, _('1) You are invited to pay the administration fee.'))
                _message = '%s\n%s' % (_message, _('2) To prepare the next sales declaration, you need to select the PV Modules you put on the market.')+' '+_('Go to the icon bar below your screen, and "Select PV Modules".'))
                _message = '%s\n\n%s' % (_message, _('IMPORTANT NOTICE'))
                _message = '%s\n\n%s' % (_message, _('You need to make this selection, before the sales period has ended.'))

                notification = nc.set_notification(_('Apply at TBS approved (%(tbs_name)s)' % {'tbs_name': tbs_member.tbs.name}), _message)

            elif new_status == PvcTbsMember.STATUS_REJECTED:
                notification = nc.set_notification(_('Apply at TBS rejected (%(tbs_name)s)' % {'tbs_name': tbs_member.tbs.name}), state['reject_reason'])
            # TODO: other statuses

            if notification is not None:
                nc.create_system_notification(notification, recipients, for_license=True)

            tbs_member.status = new_status

            self.logger.info("Updated tbs_member status [%(tbs_member_id)s], status [%(old_status)s]-> [%(new_status)s]" % {
                'tbs_member_id': tbs_member.id,
                'old_status': old_status,
                'new_status': tbs_member.status,
            })

        # this is used in case the Member was previous member
        if state.has_key('upload_contract') and state['upload_contract'] is not None:

            self.logger.info("Uploading TBS Member (%s) contract through backoffice." % tbs_member.id)
            upload_contract = UploadFileController().get_hash_temp_files(state['upload_contract'])[0]

            file_controller = AnonymousFileController(self.sa_session)
            extra = {
                'tbs_member': tbs_member,
                'type': PvcTbsMemberContractFile.TYPE_CONTRACT_FULL
            }
            iq_file = file_controller.save_file_from_tmp_for_target_group(upload_contract, 'tbs contracts', tbs_member.member.default_group, polymorphic_type=PvcTbsMemberContractFile, props=extra)
            self.sa_session.add(iq_file)
            # An uploaded contract means the status should be set to pending if it isnt set to approved.
            if old_status == new_status and tbs_member.status == PvcTbsMember.STATUS_INCOMPLETE:
                tbs_member.status = PvcTbsMember.STATUS_PENDING

        tbs_member.is_representative = state['is_representative']
        # process voting rights
        tbs_member.is_full_member = state['is_full_member']

        if state.has_key('company_registration_number'):
            tbs_member.company_registration_number = state['company_registration_number']

        if tbs_member.is_full_member:
            tbs_member.has_voting_rights = state['has_voting_rights']
        else:
            tbs_member.has_voting_rights = False

        if tbs_member.has_voting_rights:
            tbs_member.number_of_votes = state['number_of_votes']
        else:
            tbs_member.number_of_votes = None

        self.sa_session.add(tbs_member)

        # trigger new Administration fee order.
        if new_status == PvcTbsMember.STATUS_APPROVED:
            from pvcycle.order.controllers import PvcOrderController, PvcAdminFeeAlreadyExists

            oc = PvcOrderController(self.sa_session, account_id=self.get_account().id)
            try:
                order = oc.trigger_administration_fee_order(tbs_member)
            except PvcAdminFeeAlreadyExists:
                self.logger.info('Not creating administration fee, order already exists for %s' % today().year)


    def get_certificate(self, tbs_member_id, year):
        """
        gets the certificate for the given member and year
        """

        tbs_member = self._get_tbs_member(tbs_member_id)

        # check if member has a certificate
        try:
            member_certificate = self.sa_session.query(PvcTbsMemberCertificate) \
                .filter(PvcTbsMemberCertificate.tbs_member == tbs_member, PvcTbsMemberCertificate.year == year, PvcTbsMemberCertificate.status == PvcTbsMemberCertificate.STATUS_APPROVED).one()
        except NoResultFound:
            raise PublicUserException(_('No (valid) certificate found'))

        # return the tbs's certificate which contains the content
        try:
            cert = self.sa_session.query(PvcTbsCertificate) \
                .filter(PvcTbsCertificate.tbs_id == tbs_member.tbs_id, PvcTbsCertificate.year == year).one()
        except NoResultFound:
            raise Exception('No certificate content found for tbs [%s] and year [%s]' % (tbs_member.tbs_id, year))

        return cert, member_certificate

    @transaction_decorator
    def update_certificate_status(self, tbs_member_id, year, new_status):


        tbs_member = self._get_tbs_member(tbs_member_id)

        certificate = self.sa_session.query(PvcTbsMemberCertificate) \
            .filter(PvcTbsMemberCertificate.tbs_member == tbs_member, PvcTbsMemberCertificate.year == year).one()

        # operator permission check
        if self.has_permissions('pvc_tbs_license'):
            self.assert_permissions('pvc_manage_certificates')

        # tbs permission check
        else:
            self.assert_permissions("pvc_manage_certificates", license_id=certificate.tbs_member.tbs.id)

        paid = PvcTbsMemberCertificate.STATUS_PAID
        approved = PvcTbsMemberCertificate.STATUS_APPROVED
        rejected = PvcTbsMemberCertificate.STATUS_REJECTED
        revoked = PvcTbsMemberCertificate.STATUS_REVOKED

        try:
            if certificate.status == paid:
                assert new_status in (approved, rejected)
            elif certificate.status == approved:
                assert new_status in (revoked, )
            elif certificate.status == rejected:
                assert new_status in (approved, )
            elif certificate.status == revoked:
                assert new_status in (approved, )
            else:
                raise Exception('Invalid certificate status for %s' % certificate)
        except AssertionError:
            raise PublicUserException(_('Invalid new certificate status'))

        certificate.status = new_status
        self.sa_session.add(certificate)

        self.logger.info('Certificate [%s/%s] upgraded to status [%s]' % (tbs_member_id, year, new_status) )

        # send notification to participant if certificate status is approved
        if new_status == PvcTbsMemberCertificate.STATUS_APPROVED:
            nc = NotificationController(self.sa_session)
            notification = nc.set_notification(
                'A new certificate is available.',
                'A new certificate for the year %s is available for download. Look for the tab "Certificate" on your provided dashboard.' % year
            )
            nc.create_system_notification(notification, tbs_member.member.default_group)



    def validate_certificate(self, _hash):

        certificate = self.sa_session.query(PvcTbsMemberCertificate).filter(PvcTbsMemberCertificate.hash==_hash).first()

        if certificate is None:
            raise PvcCertificateNotFoundException(_('The requested certificate was not found.'))

        if certificate.status == PvcTbsMemberCertificate.STATUS_REVOKED:
            raise PvcCertificateInvalidException(_('The requested certificate has been revoked.'))

        if certificate.status == PvcTbsMemberCertificate.STATUS_REJECTED:
            raise PvcCertificateInvalidException(_('The requested certificate has been rejected.'))

        if today().year != certificate.year:
            raise PvcCertificateInvalidException(_('The requested certificate is no longer valid.'))

        if certificate.tbs_member.status != PvcTbsMember.STATUS_APPROVED:
            raise PvcCertificateInvalidException(_('The requested certificate member is currently not approved.'))

        return certificate

    @transaction_decorator
    def upgrade_to_commissioner(self, tbs_member_id):
        self.assert_permissions('license_masterdata_change_approval')

        tm = self._get_tbs_member(tbs_member_id)
        if tm.is_commissioner:
            raise PublicUserException(_('This member is already a commissioner'))

        tm.is_commissioner = True
        self.sa_session.add(tm)

        self.logger.info('Upgraded TBS Member [%s] to commissioner.' % tm.id)

        return tm


class PvcInvalidFeeException(PublicUserException): pass


