# Copyright (c) 2013 BaseSoft B.V.
# See LICENSE for details.

from basesoft import logging
from basesoft.web.contrib.license.controllers import LicenseController
from basesoft.web.core.exceptions import PublicUserException
from basesoft.web.core.forms.handler import FormValidator
from basesoft.web.core.input.validation import plugins, group_plugins
from basesoft.web.contrib.file.controllers import UploadFileController
from basesoft.web.contrib.wiki.parser import Parser
from basesoft.util.dict import dict_merge
from basesoft.web.contrib.i18n.models import Country
from basesoft.web.core.util.controllers import LicenseConfigController

from pvcycle.core.controllers import PvcTBSController, PvcAccountantController, PvcMemberController, UniqueVATNumber,\
    ServiceLevelReplacementFieldsContext, FrRCSLocationValidation
from pvcycle.core.models import PvcMemberLicense, PvcTbsMember

logger = logging.getLogger("iqportal.iqreport.core")

class DefaultPvcHandler:

    def __init__(self, request, **kwargs):

        self._request = request
        self._sa_session = request.Session()
        self._account_id = request.session['user_id']

    def validate_company(self, args, kwargs=None):

        country_map = {}
        for c in self._sa_session.query(Country).all():
            country_map[c.iso_3166_a2] = c.id

        unique_vat_args = {'db_session': self._sa_session}
        if kwargs is not None:
            unique_vat_args = dict(unique_vat_args.items()+kwargs.items())


        behaviour = {
            'name': [(plugins.Required, {})],
            'vat_no': [(plugins.Required, {}), (plugins.VATOffline, {}), (UniqueVATNumber, unique_vat_args)],
        }

        country_id = int(kwargs['country_id']) if kwargs.has_key('country_id') else None

        if country_id == country_map['NL']: # netherlands
            behaviour['chamber_of_commerce_nr'] = [ (plugins.Required, {}), ]

        elif country_id == country_map['ES']: # spain
            behaviour['es_cif_nr'] = [(plugins.Required, {}), ]

        elif country_id == country_map['FR']: # france
            behaviour['fr_rcs_number'] = [(plugins.Required, {}), (plugins.Number, {})  ]
            behaviour['fr_rcs_location'] = [(plugins.Required, {}), (FrRCSLocationValidation, {}) ]
            behaviour['fr_company_type'] = [(plugins.Required, {}), ]

        elif country_id == country_map['IT']: # italy
            behaviour['it_fiscal_code'] = [(plugins.Required, {}), ]

        return FormValidator(behaviour).validate(args)

    def validate_company_non_eu(self, args, kwargs=None):

        unique_vat_args = {'db_session': self._sa_session}
        if kwargs is not None:
            unique_vat_args = dict(unique_vat_args.items()+kwargs.items())

        form_handler = FormValidator({
            'name': [(plugins.Required, {})],
            'vat_no': [(plugins.VATOffline, {}), (UniqueVATNumber, unique_vat_args)], # VAT is not required for non-eu countries
        })
        return form_handler.validate(args)

    def validate_address(self, args):
        country_map = {}
        country_id = None

        for c in self._sa_session.query(Country).all():
            country_map[c.iso_3166_a2] = c.id

        if args.has_key('country') and args['country'] not in [None, '']:
            country_id = int(args['country'])

        form_handler = {
            'street': [(plugins.Required, {})],
            'postcode': [(plugins.Required, {})],
            'city': [(plugins.Required, {})],
            'country': [(plugins.Required, {})],
        }

        if country_id == country_map['IT']: # italy
            form_handler['state'] = [(plugins.Required, {})]

        return FormValidator(form_handler).validate(args)

    def validate_bank_account_non_eu(self, args):
        form_handler = FormValidator({
            'bic': [(plugins.SWIFTBIC, {'db_session': self._sa_session})],
        })
        return form_handler.validate(args)

    def validate_bank_account(self, args):
        form_handler = FormValidator({
            'bic': [(plugins.Required, {}), (plugins.SWIFTBIC, {'db_session': self._sa_session})],
            'iban': [(plugins.Required, {}), (plugins.IBAN, {})],
        })
        return form_handler.validate(args)

    def validate_person(self, args):
        form_handler = FormValidator({
            'first_name': [(plugins.Required, {})],
            'last_name': [(plugins.Required, {})],
            'gender': [(plugins.Required, {})],
            'preferred_locale': [(plugins.Required, {})],
            'email': [(plugins.Required, {}), (plugins.Email, {})],
            'phone': [(plugins.Required, {})],
        })
        return form_handler.validate(args)

class AccountantHandler:

    def __init__(self, request, **kwargs):

        self._request = request
        self._sa_session = request.Session()
        self._account_id = request.session['user_id']

        self._controller = PvcAccountantController(self._sa_session, account_id=self._account_id, request=request)

    def load_accountant(self, license_id):
        state = self._controller.load_accountant(license_id)
        return state

    def save_accountant(self, state):

        license_ = self._controller.save_accountant(state)
        return license_.id

class TbsHandler:

    def __init__(self, request, **kwargs):
        self._request = request
        self._sa_session = request.Session()
        self._account_id = request.session['user_id']

        self._controller = PvcTBSController(self._sa_session, account_id=self._account_id, request=request)

    def validate_tbs_information(self, args):
        form_handler = FormValidator({
            'tbs_type': [(plugins.Required, {})],
        })
        return form_handler.validate(args)

    def load_tbs(self, license_id):
        state = self._controller.load_tbs(license_id)
        return state

    def save_tbs(self, state):
        license_ = self._controller.save_tbs(state)
        self._controller.refresh_roe_countries_for_pv_modules()
        return license_.id

    def save_tbs_settings(self, _id, state):
        orig = self._controller.load_tbs(_id) # expects state to be filled completly
        s = dict_merge(orig, state)
        s['dbID'] = _id
        license_ = self._controller.save_tbs(s)
        return license_.id

    def get_take_back_schemes(self):
        return [{'value': t.id, 'label': t.name} for t in self._controller.get_take_back_schemes()]

    def get_take_back_schemes_for_service_levels(self):
        return [{'value': t.id, 'label': t.name} for t in self._controller.get_take_back_schemes(permissions=('pvc_tbs_service_level', ) ) if not self._has_membership(t) ]

    def get_take_back_schemes_for_membership(self):
        return [{'value': t.id, 'label': t.name} for t in self._controller.get_take_back_schemes(permissions=('pvc_tbs_membership_level', ) ) if self._has_membership(t)  ]

    def _has_membership(self, _license):
        lcc = LicenseConfigController(self._sa_session, scope=_license)
        return lcc.get('pvc.tbs.use.membership.instead.of.service.level')


    def load_service_level(self, _id):
        return self._controller.get_service_level_state(_id)

    def load_membership(self, _id):
        return self._controller.get_membership_state(_id)

    def validate_save_service_level(self, args):
        form_handler = FormValidator({
            'tbs_id': [(plugins.Required, {})],
            'contract_text': [(plugins.Required, {})],
            'name': [(plugins.Required, {})],
            'start_date': [(plugins.Required, {})],
            'start_year': [(plugins.Required, {})],
            'fee': [(plugins.Required, {})],

        })
        return form_handler.validate(args)

    def validate_save_membership(self, args):
        form_handler = FormValidator({
            'tbs_id': [(plugins.Required, {})],
            'name': [(plugins.Required, {})],
            'start_year': [(plugins.Required, {})],
            'fee': [(plugins.Required, {})],

        })
        return form_handler.validate(args)

    def save_service_level(self, _id, state):
        return self._controller.save_service_level_state(_id, state).id

    def update_active_service_level_fees(self, _id, contribution_fee, administration_fee):
        a, c = self._controller.update_active_service_level_fees(_id, contribution_fee, administration_fee)
        return [a, c if c is None else len(c)]

    def save_membership(self, _id, state):
        return self._controller.save_membership_state(_id, state).id

    def delete_service_levels(self, ids):
        for _id in ids:
            self._controller.delete_service_level(_id)

    def get_contract_replacement_fields(self):
        return self._controller.get_contract_replacement_fields()

    def get_sl_replacement_fields(self):
        return self._controller.get_sl_replacement_fields()

    # certificate
    def load_tbs_certificate(self, _id):
        return self._controller.get_tbs_certificate_state(_id)

    def validate_save_tbs_certificate(self, args):
        form_handler = FormValidator({
            'tbs_id': [(plugins.Required, {})],
            'year': [(plugins.Required, {})],
            'text': [(plugins.Required, {})],

        })
        return form_handler.validate(args)

    def save_tbs_certificate(self, id_, state):
        return self._controller.save_tbs_certificate_state(id_, state).id

    def delete_tbs_certificates(self, ids):
        for _id in ids:
            self._controller.delete_tbs_certificate(_id)

    def get_certificate_replacement_fields(self):
        return self._controller.get_certificate_replacement_fields()

    def save_tbs_texts(self, tbs_id, state):
        self._controller.save_tbs_texts(tbs_id, state)
        return True

    def load_tbs_texts(self, tbs_id):
        state = self._controller.load_tbs(tbs_id)
        return state['tbs']

    def load_roe_countries(self, _id=None):
        # using details panel, supplies an ID => bitbucket
        return {'countries': [c.id for c in self._controller.get_roe_countries()]}

    def save_roe_countries(self, _id, values):
        rv = self._controller.save_roe_countries(values['countries'])
        if not rv:
            return False

        self._controller.refresh_roe_countries_for_pv_modules()

        return True

    def get_available_tbs_countries(self):
        return [{'value': c.id, 'label': c.name} for c in self._controller.get_available_tbs_countries()]


class MemberHandler:

    def __init__(self, request, **kwargs):

        self._request = request
        self._sa_session = request.Session()
        self._account_id = request.session['user_id']

        self._controller = PvcMemberController(self._sa_session, account_id=self._account_id, request=request)

    def post_login(self):
        """
        post login is called after every login of a member
        """

        rv = {
        }

        lc = LicenseController(self._sa_session, account_id = self._account_id, request=self._request)
        license_id = lc.get_default_group().license.id

        if not self._controller.is_applied_at_tbs(license_id):
            rv['show_wizard_for_license'] = license_id

        # show_apply_tbs_panels is a list of tuples (PvcMember.id, PvcMemberLicense.id)
        rv['show_apply_tbs_panels'] = [
            (x.id, x.member.id) for x in self._controller.get_unfinished_tbs_members(license_id)]

        return rv

    def load_member_license(self, license_id):
        state = self._controller.load_member_license(license_id)
        return state

    def save_member_license(self, state):
        license_ = self._controller.save_member_license(state)
        return license_.id

    def get_take_back_schemes(self, member_id=None, filter_applied=False, _type=None):
        """
        returns tbs's where the (new) member can apply to
        """
        rv = [{'label': x.name, 'value': x.id} for x in self._controller.get_take_back_schemes(_type=_type)]
        if not filter_applied:
            return rv

        applied = [m.tbs.id for m in self._controller.get_tbs_members(member_id)]
        return [x for x in rv if x['value'] not in applied]

    def get_active_for_tbs(self, member_id=None, is_producer=None, is_commissioner=None):
        return [{'label': tm.tbs.name, 'value': tm.tbs.id} for tm in self._controller.get_active_for_tbs(member_id, is_producer, is_commissioner)]

    def get_available_countries(self, member_id=None, tbs_member_id=None):
        """
        returns a list of countries which are available for the different Take Back Schemes the supplied member has access to.
        """
        return [{'label': x.name, 'value': x.id} for x in self._controller.get_member_countries(member_id, tbs_member_id=tbs_member_id)]

    def get_available_countries_for_member_at_tbs(self, member_id, tbs_id):
        if tbs_id is None or tbs_id == '':
            return []
        mc = PvcMemberController(self._sa_session, account_id=self._account_id, request=self._request)
        tbs_member = [tm for tm in mc.get_active_for_tbs(member_id) if tm.tbs.id==int(tbs_id)].pop()
        return [{'label': x.name, 'value': x.id} for x in self._controller.get_tbs_member_countries(tbs_member)]

    def get_service_levels_select(self, tbs_id):
        """
        returns the service levels for the given tbs_id
        """
        return [{'label': x.name, 'value': x.id} for x in self._controller.get_service_levels(tbs_id)]

    def get_service_levels(self, tbs_id, license_id):
        """
        returns the service levels for the given tbs_id
        """

        license_ = self._controller.get_licenses(license_id)[0]

        return [{
            'id': sl.id,
            'name': sl.name,
            'contract_annex_text': Parser().render_to_html(sl.contract_annex_text, ServiceLevelReplacementFieldsContext(member=license_, service_level=sl)),

        } for sl in self._controller.get_service_levels(tbs_id)]


    def get_membership_select(self, tbs_id):
        return [{'label': x.name, 'value': x.id} for x in self._controller.get_memberships(tbs_id)]

    def validate_apply_at_tbs(self, args):

        validation = {
            'tbs_id': [(plugins.Required, {})],
            'license_id': [(plugins.Required, {})],
        }
        group_validation = []

        if not args.has_key('tbs_id') or not args['tbs_id']:
            return FormValidator(validation).validate(args)

        # get the license configs for this TBS to see what all is required.
        config = self.get_tbs_configs(args['tbs_id'])
        if config['pvc.tbs.require.company.registration.number']:
            validation['company_registration_number'] = [(plugins.Required, {})]

        if config['pvc.tbs.use.membership.instead.of.service.level']:
            validation['membership_id'] = [(plugins.Required, {})]
        else:

            group_validation += [
                (group_plugins.Required, {
                    'is_commissioner': {'is_producer': False },
                    'is_producer': {'is_commissioner': False }
                }),
                # service level is required if producer is checked
                (group_plugins.Required, {'service_level_id': {'is_producer': True} }),
                # upload_contract is required if producer is checked and upload_later is unchecked
                (group_plugins.Required, {'upload_contract': {'is_producer': True, 'upload_later': False} }),
            ]

        if config['pvc.tbs.require.accepting.of.terms.and.conditions']:
            validation['terms_and_conditions'] = [(plugins.RequiredCheckbox, {})]

        form_handler = FormValidator(validation, group_plugins = group_validation)
        return form_handler.validate(args)

    def apply_at_tbs(self, member_license_id, tbs_member_id, form_values):

        service_level_id = None
        if form_values['service_level_id'] != None and form_values['service_level_id'] != '':
            service_level_id = int(form_values['service_level_id'])

        membership_id = None
        if form_values['membership_id'] != None and form_values['membership_id'] != '':
            membership_id = int(form_values['membership_id'])

        accepted_tac = form_values['terms_and_conditions']

        tbs_license_id = int(form_values['tbs_id'])
        is_commissioner = form_values['is_commissioner']
        is_producer = form_values['is_producer']
        company_registration_number = None if form_values['company_registration_number'] == '' else form_values['company_registration_number']
        was_member_in_2013 = False
        if form_values.has_key('was_member_in_2013'):
            was_member_in_2013 = True if form_values['was_member_in_2013'] == 'true' else False

        if not is_producer:
            service_level_id = None
            form_values['upload_contract'] = None

        upload_contract = None
        if service_level_id is not None and not form_values['upload_later']:
            upload_contract = UploadFileController().get_hash_temp_files(form_values['upload_contract'])[0]

        tbs_member = self._controller.apply_at_tbs(tbs_license_id, member_license_id, tbs_member_id, service_level_id, is_commissioner,
                                      is_producer, upload_contract, was_member_in_2013, membership_id=membership_id,
                                      company_reg_no=company_registration_number, accepted_tac=accepted_tac)

        return tbs_member.status == PvcTbsMember.STATUS_APPROVED

    def get_license_master_data_state(self, _id):
        return self._controller.load_master_data_state(_id)

    def save_license_master_data_state(self, _id, state):
        self._controller.update_master_data(_id, state)
        return _id

    def load_tbs_member(self, tbs_member_id):
        return self._controller.load_tbs_member(tbs_member_id)

    def get_contract(self, tbs_member_id):
        return self._controller.get_contract(tbs_member_id)

    def validate_save_tbs_member(self, args):

        form_handler = FormValidator({
            'status': [(plugins.Required, {})],
            'number_of_votes': [(plugins.Number, {})],
        },
        group_plugins = [
            (group_plugins.Required, {'number_of_votes': {'has_voting_rights': True} }),
        ])

        return form_handler.validate(args)

    def save_tbs_member(self, tbs_member_id, state):
        return self._controller.save_tbs_member(tbs_member_id, state)


    def get_member_licenses(self):
        """
        returns a list of member licenses this account can register to a tbs for
        """
        return [{'value': x.id, 'label': x.name} for x in self._controller.get_licenses()]

    def get_eu_countries_no_vat(self):
        return ([c.id for c in self._controller.get_eu_countries()], [c.id for c in self._controller.get_eu_countries_required_vat()], )

    def get_eu_countries(self):
        return [c.id for c in self._controller.get_eu_countries()]

    def get_efta_countries(self):
        return [c.id for c in self._controller.get_efta_countries()]

    def get_tbs_memberships_in_2013(self, tbs_id, registration=False):

        if registration:
            tbs = self._controller.get_take_back_schemes(tbs_id)
        else:
            tbs = self._controller._get_take_back_schemes(tbs_id).pop()
        lcc = LicenseConfigController(self._sa_session, scope=tbs)
        return lcc.get('pvc.tbs.was.active.in.2013')

    def get_tbs_configs(self, tbs_id):
        # Only return the PVC configs
        tbs = self._controller.get_take_back_schemes(tbs_id)
        lcc = LicenseConfigController(self._sa_session, scope=tbs)
        rv = {}
        for k, v in lcc.get_configs().iteritems():
            if k[:3] != 'pvc':
                continue
            rv[k] = v
        return rv

    def is_pvc_member(self):
        _license = self._controller.get_default_group().license
        return {'is_pvc_member': isinstance(_license, PvcMemberLicense), 'license_id': _license.id}


    def get_tac(self, tbs_id):
        tbs = self._controller.get_take_back_schemes(tbs_id)
        p = Parser()
        return p.render_to_html(tbs.terms_and_conditions_text)

    def upgrade_to_commissioner(self, tbs_member_id):
        self._controller.upgrade_to_commissioner(tbs_member_id)
        return True

    def update_certificate_status(self, tbs_member_id, year, new_status):
        return self._controller.update_certificate_status(tbs_member_id, year, new_status)

class PublicPageHandler:

    def __init__(self, request, **kwargs):

        self._request = request
        self._sa_session = request.Session()

    def verify_hash(self, _hash):
        c = PvcMemberController(self._sa_session)
        certificate = c.validate_certificate(_hash)
        return {
            'year': certificate.year,
            'issue_date': certificate.issue_date,
            'member': certificate.tbs_member.member.name,
            'tbs': certificate.tbs_member.tbs.name,
        }

    def validate_certificate(self, args):
        form_handler = FormValidator({
            'certificate_hash': [(plugins.Required, {})],
        })
        return form_handler.validate(args)

handlers = (DefaultPvcHandler, AccountantHandler, TbsHandler, MemberHandler, PublicPageHandler)


