'''
Created on Oct 8, 2013

@author: marten
'''
from django.utils.translation import ugettext as _

from basesoft.web.core.exceptions import SystemException, PublicUserException, SecurityException,\
    PublicSecurityException
from basesoft.web.contrib.order.controllers import AbstractOrderController

from pvcycle.core.controllers import AbstractPvcController, PvcNotificationController
from pvcycle.order.models import PvcOrder, PvcOrderLineAdministrationFee, PvcDeclarationOrder, PvcDeclarationOrderLine,\
    PvcOrderLineMembershipFee, PvcDeclarationReminderMail, PvcDeclarationOrderLineMinimumCF
from pvcycle.core.models import PvcTbsMember, PvcTbsLicense,\
    PvcTbsMemberCertificate, PvcOperatorLicense
from basesoft.web.core.util.controllers import LicenseConfigController, transaction_decorator
from sqlalchemy.sql.expression import and_, extract
from pvcycle.pvmodule.models import PvcPVModule, PvcTbsMemberPVModule

import datetime
from basesoft.util import virtual_date
from pvcycle.core.exceptions import PvcPVModuleNotApprovedException
from dateutil.relativedelta import relativedelta
from basesoft.util.virtual_date import today, now
from basesoft.web.contrib.order.models import OrderPaymentStatus
from basesoft.web.contrib.notification.controllers import NotificationController
from pvcycle.pvmodule.controllers import PvcPVModuleController
from pvcycle.subrogation.models import PvcSubrogation
from iqportal.iqube.core.controllers import IQubeAggregator

class PvcOrderController(AbstractOrderController, AbstractPvcController):
    _instance_class = PvcOrder

    def set_additional_fields(self):
        self._set_additional_fields({
            self._instance_class: ['tbs_member']
        }, {
            PvcOrderLineAdministrationFee: ['tbs_member_service_level'],
            PvcOrderLineMembershipFee: ['tbs_member_membership'],
        })

    @transaction_decorator
    def trigger_administration_fee_order(self, tbs_member):
        """
        This is triggered on registration of a new TBS Member; OR in case the member upgrades their Service Level.
        """

        assert isinstance(tbs_member, PvcTbsMember)

        if tbs_member.status != PvcTbsMember.STATUS_APPROVED:
            raise SystemException('Trying to trigger an Administration/Membership Order for an unapproved TBS Member')

        # we get if this TBS uses Memberships instead of Service Levels
        lcc = LicenseConfigController(self.sa_session, scope=tbs_member.tbs)
        use_membership = lcc.get('pvc.tbs.use.membership.instead.of.service.level')

        active_level = None
        if use_membership:
            fee_title = _('Membership Fee')
            type_title = _('Membership')
            fl = PvcOrderLineMembershipFee # Memberships use another ORDER line type.
            active_level = tbs_member.active_membership
            years = lcc.get('pvc.tbs.membership.valid.for.years')
            year_check = extract('year', PvcOrder.display_date) + years > today().year

        else:
            years = 1
            fee_title = _('Administration Fee')
            type_title = _('Service Level')
            fl = PvcOrderLineAdministrationFee
            active_level = tbs_member.active_service_level
            year_check = extract('year', PvcOrder.display_date) == today().year
            if active_level is None and not tbs_member.is_producer:
                return None # COMMISSIONERS do not have Service Levels.

        if active_level is None:
            raise PublicUserException(_('Cannot approve a member who has no active %(membership_type)s.' % {'membership_type': type_title}))

        current_fee = getattr(active_level, active_level.BACKREF).get_current_administration_fee()
        if current_fee is None:
            raise PublicUserException(_('Cannot approve a member without an %(fee_type)s set on %(membership_type)s: %(membership_name)s.' % {
                'fee_type' : fee_title,
                'membership_type': type_title,
                'membership_name': getattr(active_level, active_level.BACKREF).name
            }))

        # check if the member already has a AF for this year (or period for memberships)
        q = self.sa_session.query(PvcOrder).join(fl) \
            .filter(
                year_check,
                getattr(fl, fl.BACKREF)==active_level,
            )
        if (len(q.all()) > 0):
            raise PvcAdminFeeAlreadyExists('%(fee_type)s already exists for year %(year)s' % {'fee_type': fee_title, 'year': today().year})

        period = getattr(active_level, active_level.BACKREF).get_current_fee_start_year()
        if years > 1:
            period = '%s - %s' % (str(period), str(period+years-1))
            description = _('%(fee_type)s for period %(period)s' % {'fee_type': fee_title, 'period': period})
        else:
            description = _('%(fee_type)s for year %(year)s' % {'fee_type': fee_title, 'year': period})

        # create the AF order
        lines = [{
            'polymorphic_type': fl,
            'description': description,
            'quantity': 1,
            'price': current_fee,
            fl.BACKREF: active_level,
        }]

        self.logger.info('trigger_administration_fee_order: Created new Administration/Membership Fee Order for TBS Member [%s]' % tbs_member.id)
        order = self._save_or_update_order(None, lines, merchant=tbs_member.tbs, scope=tbs_member.member.default_group, tbs_member=tbs_member, display_date=now())

        # if the new order has 0 amount, trigger the certificate directly
        if order is not None and order.invoiced_amount == 0:
            self.create_certificate(order)

        return order


    def _is_administrative_fee_order(self, order):

        # check if its a Administrative/Membership Fee order
        is_af_order = False
        for line in order.lines:
            if line.polymorphic_type == 'PvcOrderLineAdministrationFee':
                is_af_order = True
                break
            elif line.polymorphic_type == 'PvcOrderLineMembershipFee':
                is_af_order = True
                break

        return is_af_order

    @transaction_decorator
    def trigger_yearly_administrative_fees(self, tbs):
        """
        Called by the pvcycle YearlyOrderTask to trigger new AF / Membership orders
        """

        lcc = LicenseConfigController(self.sa_session, scope=tbs)
        fee_day = lcc.get('pvc.tbs.fee.day')
        if today().month != 1 or today().day != fee_day:
            self.logger.info('today (%s) is not the pvc.tbs.fee.day (%s)' % (today(), fee_day))
            return False


        # query all TbsMembers that are supposed to receive a new AF / Membership order (status STATUS_APPROVED)
        tbs_members = self.sa_session.query(PvcTbsMember) \
            .filter(PvcTbsMember.tbs == tbs) \
            .filter(PvcTbsMember.status == PvcTbsMember.STATUS_APPROVED) \
            .all()

        for tbs_member in tbs_members:
            try:
                self.trigger_administration_fee_order(tbs_member)

            # its possible that a fee already exists
            # - the task is run twice
            # - membership fee's are only created once per 'pvc.tbs.membership.valid.for.years'
            except PvcAdminFeeAlreadyExists, e:
                self.logger.info("trigger_yearly_administrative_fees: skipping order for tbs_member [%s] because [%s]" % (tbs_member.id, e.message))

            # ignore public user exceptions
            # - member does not have active server/membership level
            except PublicUserException, e:
                self.logger.info("trigger_yearly_administrative_fees: skipping order for tbs_member [%s] because [%s]" % (tbs_member.id, e.message))

        # secondly, update everyone that is STATUS_RESIGNING to STATUS_RESIGNED
        tbs_members = self.sa_session.query(PvcTbsMember) \
            .filter(PvcTbsMember.tbs == tbs) \
            .filter(PvcTbsMember.status == PvcTbsMember.STATUS_RESIGNING) \
            .all()
        for tbs_member in tbs_members:
            tbs_member.status = PvcTbsMember.STATUS_RESIGNED
            self.logger.info("trigger_yearly_administrative_fees: updated tbs_member [%s] to status resigned" % tbs_member)

        return True

    @transaction_decorator
    def payment_status_change_cb(self, order):
        """
        Called by the payment controller when the order payment changes.
        """

        if order.get_current_payment_status() == OrderPaymentStatus.PAYMENT_COMPLETED:

            if self._is_administrative_fee_order(order):
                self.create_certificate(order)

    def create_certificate(self, order):
        """
        Gives the owner of the order a certificate. Called by payment_status_change_cb() or when the order amount is 0.
        """
        assert self._is_administrative_fee_order(order)

        # create a new certificate for this member
        tbs_member = self.sa_session.query(PvcTbsMember) \
            .filter(PvcTbsMember.tbs == order.merchant, PvcTbsMember.member == order.client).one()

        year = order.display_date.year

        cert = PvcTbsMemberCertificate()
        cert.year = year
        cert.tbs_member = tbs_member
        cert.status = PvcTbsMemberCertificate.STATUS_PAID
        cert.hash = self.generate_unique_hash()
        self.sa_session.add(cert)

        self.logger.info('payment_status_change_cb: Gave PVC certificate to tbs_member_id [%s] and year [%s]' % (tbs_member.id, year) )


    def generate_unique_hash(self):
        """
        Generates a unique hash.
        """
        import random
        from hashlib import sha256 as sha
        while True:
            _hash = sha(str(random.random())).hexdigest()[:8]
            if self.sa_session.query(PvcTbsMemberCertificate).filter(PvcTbsMemberCertificate.hash == _hash).count() == 0:
                break
        return _hash


class PvcAdminFeeAlreadyExists(PublicUserException): pass


class PvcDeclarationOrderController(AbstractOrderController, AbstractPvcController):
    _instance_class = PvcDeclarationOrder

    merchant_uses_weight = None

    def _get_original_order(self, order):
        if order.parent_order is None:
            return order
        return self._get_original_order(order.parent_order)

    def _assert_valid_line_removal(self, order, line):
        super(PvcDeclarationOrderController, self)._assert_valid_line_removal(order, line)
        # rejected by commissioner is valid for removal.
        if line.tbs_member_module.subrogation is not None and line.tbs_member_module.commissioner_approval_status == PvcTbsMemberPVModule.COMMISSIONER_APPROVAL_STATUS_REJECTED:
            return
        if line.tbs_member_module.module.approval_status == PvcPVModule.PV_MODULE_STATUS_APPROVED:
            raise PublicUserException(_('You cannot remove a line with an approved PV Module. Please declare 0 instead and deactive from the PV Module overview for future declarations.'))

    def set_additional_fields(self):
        self._set_additional_fields({
            self._instance_class: ['tbs_member', 'year', 'month', 'quarter', 'producer']
        }, {
            PvcDeclarationOrderLine: ['tbs_member_module'],
        })

    def load_state(self, _id, subrogations=False):
        order = self.get(_id).pop()

        field = None

        lcc = LicenseConfigController(self.sa_session, scope=order.merchant)
        if lcc.get('pvc.tbs.declaration.use.weight.for.amount.calculations'):
            field = 'weight'

        rv = super(PvcDeclarationOrderController, self).load_state(order)
        rv['year'] = order.year
        rv['month'] = order.month
        rv['quarter'] = order.quarter
        rv['amount_calculation_field'] = field
        rv['tbs_member_id'] = order.tbs_member.id
        rv['producer_id'] = order.producer.id
        if subrogations:
            rv['subrogations'] = [{'value': s.id, 'label': s.commissioner.member.name} for s in self.get_available_subrogations_for_order(_id)]
            rv['can_pay'] = len(self.get(_id, priv=PvcDeclarationOrder.PRIV_PAY)) == 1
            rv['can_declare'] = len(self.get(_id, priv=PvcDeclarationOrder.PRIV_UPDATE)) == 1

        return rv

    def _load_lines_state(self, lines, for_export=False):
        if for_export:
            return self.__load_lines_state_for_export(lines)
        rv = []
        additional = []
        for line in lines:
            if isinstance(line, PvcDeclarationOrderLineMinimumCF):
                additional = super(PvcDeclarationOrderController, self)._load_lines_state([line])
                continue
            module = line.tbs_member_module.module
            rv.append({
                'id': line.id,
                'index': line.index,
                'description': line.description,
                'price_per_unit': line.price_per_unit,
                'quantity': line.quantity,
                'vat_rate': line.vat_rate,
                'gl_account': line.gl_account_number,
                'line_amount_excluding_vat': line.line_amount_excluding_vat,
                'line_amount_including_vat': line.line_amount_including_vat,
                'supplier': module.supplier.name,
                'type': module.type.name,
                'cell_type': module.cell_type.name,
                'nominal_power': module.nominal_power,
                'weight': module.weight,
                'country': line.tbs_member_module.country.name,
                'surface': module.width*module.length,
                'module_status': module.approval_status,
                'vat_rate': line.vat_rate,
            })

        rv = sorted(rv, key=lambda k: (k['country'], k['index']))
        return rv + additional

    def _load_state_for_export(self, order):
        rv = super(PvcDeclarationOrderController, self)._load_state_for_export(order)
        rv['properties'] = [{
            'key': 'year',
            'value_int': order.year
        },{
            'key': 'month',
            'value_int': order.month
        },{
            'key': 'quarter',
            'value_int': order.quarter
        }]
        return rv

    def __load_lines_state_for_export(self, lines):
        rv = []
        for line in lines:
            if isinstance(line, PvcDeclarationOrderLineMinimumCF):
                rv += super(PvcDeclarationOrderController, self)._load_lines_state([line])
                continue

            module = line.tbs_member_module.module
            rv.append({
                'index': line.index,
                'polymorphic_type': line.polymorphic_type,
                'description': line.description,
                'price_per_unit': line.price_per_unit,
                'quantity': line.quantity,
                'vat_rate': line.vat_rate,
                'gl_account': line.gl_account_number,
                'line_amount_excluding_vat': line.line_amount_excluding_vat,
                'line_amount_including_vat': line.line_amount_including_vat,
                'properties': [{
                        'key': 'supplier',
                        'value_str': module.supplier.name,
                    },{
                        'key': 'type',
                        'value_str': module.type.name,
                    },{
                        'key': 'cell_type',
                        'value_str': module.cell_type.name,
                    },{
                        'key': 'nominal_power',
                        'value_decimal': module.nominal_power,
                    },{
                        'key': 'weight',
                        'value_decimal': module.weight,
                    },{
                        'key': 'country_name',
                        'value_str': line.tbs_member_module.country.name,
                    },{
                        'key': 'country_iso_3166_a2',
                        'value_str': line.tbs_member_module.country.iso_3166_a2,
                    },{
                        'key': 'country_iso_3166_a3',
                        'value_str': line.tbs_member_module.country.iso_3166_a3,
                    },{
                        'key': 'surface',
                        'value_decimal': module.width*module.length,
                    }],
            })

        return rv

    @transaction_decorator
    def _subrogate_order(self, order, quantities, subrogate, additional_lines=None):
        orders = {}

        line_modules = {}
        for line in order.lines:
            line_modules['line_%s' % line.index] = line

        for line, subrogation_id in subrogate.iteritems():
            if not subrogation_id in [s.id for s in order.tbs_member.subrogations]:
                raise PublicSecurityException(_('You do not have access to the supplied Subrogation.'))

            subrogation = [s for s in order.tbs_member.subrogations if s.id == subrogation_id].pop()

            if not subrogation.status == PvcSubrogation.STATUS_ACTIVE:
                raise PublicSecurityException(_('The supplied Subrogation is not active.'))

            if subrogation.commissioner is None:
                self.logger.warn('Approved subrogation without a commissioner, something went wrong, order: %s, subrogation: %s' % (order.id, subrogation.id))
                raise SecurityException('Approved subrogation without a commissioner, something went wrong!')

            if not orders.has_key(subrogation.id):
                orders[subrogation.id] = {'tbs_member': subrogation.commissioner, 'producer': order.tbs_member, 'lines': []}

            orders[subrogation.id]['lines'].append({
                'polymorphic_type': PvcDeclarationOrderLine,
                'description': '%s' % line_modules[line].description,
                'quantity': int(quantities[line]),
                'price': line_modules[line].price_per_unit,
                'tbs_member_module': self.__get_pv_modules_for_subrogation(line_modules[line].tbs_member_module, subrogation)
            })
            del quantities[line]

        fields = self._additional_order_fields[self._instance_class]
        kwargs = {}
        for f in fields:
            kwargs[f] = getattr(order, f)

        rv = []
        for subrogation_id, local_order in orders.iteritems():
            kwargs['producer'] = local_order['producer']
            kwargs['tbs_member'] = local_order['tbs_member']

            if not self._is_undeclared_period(**kwargs):
                raise PublicUserException(_('A declaration for this period already exists at commissioner %(commissioner)s, please use the resubmit option.' % {'commissioner': kwargs['tbs_member'].member.name}))

            o = self._save_or_update_order(None, local_order['lines'], merchant=order.merchant, scope=local_order['tbs_member'].member.default_group, **kwargs)
            o.status = self._instance_class.STATUS_PENDING
            self.sa_session.add(o)
            rv.append(o)

            self.sa_session.add(o)

        # remove lines from original order.
        if len(quantities) == 0:
            # all lines are moved to subrogations.
            self.sa_session.delete(order)
            self.sa_session.flush()
            return rv

        lines = []
        for line, quantity in quantities.iteritems():
            lines.append({
                'polymorphic_type': PvcDeclarationOrderLine,
                'description': '%s' % line_modules[line].description,
                'quantity': int(quantity),
                'price': line_modules[line].price_per_unit,
                'tbs_member_module': line_modules[line].tbs_member_module,
            })
        order = self._save_or_update_order(order.id, lines, merchant=order.tbs_member.tbs, scope=order.tbs_member.member.default_group, \
                                          tbs_member=order.tbs_member, year=order.year, month=order.month, quarter=order.quarter, producer=order.producer)
        order.status = self._instance_class.STATUS_PENDING
        self.sa_session.add(order)

        order = self.update_declaration(order.id, quantities, additional_lines=additional_lines)

        rv.append(order)

        self.sa_session.flush()
        return rv

    def __get_pv_modules_for_subrogation(self, module, subrogation):
        """
        Get or create the PV Module supplied for this subrogation.
        """
        c = PvcPVModuleController(self.sa_session, account_id=self.get_account().id, request=self.request)
        return c.save_or_update_module_for_tbs_member(module.tbs_member, module.country, module.module, subrogation_id=subrogation)

    @transaction_decorator
    def save_default_privilege_object_group(self, order, target_group):
        """
        Override this method in order to assign privs based on Subrogation.
        """
        # normal behavior, no subrogation
        if order.tbs_member == order.producer:
            return super(PvcDeclarationOrderController, self).save_default_privilege_object_group(order, target_group=target_group)

        # give producer the right privs.
        super(PvcDeclarationOrderController, self).save_default_privilege_object_group(order, target_group=order.producer.member.default_group, privileges=order.producer_privileges)

        # give the commissioner the right privs.
        self._grant_or_revoke(order, privileges=order.tbs_member_privileges, grant=True, group=True, target_group=order.tbs_member.member.default_group, check_own_privs=False)

    @transaction_decorator
    def update_declaration(self, _id, quantities, additional_lines=None, subrogate=None):
        self.assert_permissions('pvc_declare_declarations')
        order = self.get(_id, priv=self._instance_class.PRIV_UPDATE).pop()
        if order.status == self._instance_class.STATUS_CREDITED:
            raise PublicSecurityException(_('You cannot update an already Credited order.'))

        if order.status not in [self._instance_class.STATUS_NEW, self._instance_class.STATUS_PENDING]:
            if order.resubmit_request != self._instance_class.RESUBMIT_REQUEST_APPROVED:
                raise PublicUserException(_('You cannot resubmit this order.'))
            return self.resubmit_declaration(order, quantities, additional_lines) # TODO: subrogate

        if order.type == self._instance_class.TYPE_CREDIT:
            raise PublicSecurityException(_('You cannot update a Credit Order.'))

        # We are dealing with a normal Declare (or Redeclare)

        if len(quantities) < len(order.lines):
            raise PublicUserException(_('Number of lines for Order #%s does not match with supplied quantities.' % order.id))

        if subrogate is not None:
            if order.tbs_member != order.producer:
                raise PublicUserException(_('Cannot subrogate an Order which is already subrogated.'))

            return self._subrogate_order(order, quantities, subrogate, additional_lines)

        lines = []
        used_modules = []
        for line in order.lines:
            quantity = int(quantities['line_%s' % line.index])
            if quantity < 0:
                raise PublicUserException(_('All quantities must be larger or equal to zero.'))

            lines.append({
                'polymorphic_type': PvcDeclarationOrderLine,
                'description': '%s' % line.description,
                'quantity': quantity,
                'price': line.price_per_unit,
                'tbs_member_module': line.tbs_member_module
            })
            used_modules.append(line.tbs_member_module.id)

        # Additional lines as per ticket #156
        if additional_lines is not None and len(additional_lines) > 0:
            af = self._get_price(order.producer)

            possible = {}
            for m in order.producer.modules:
                if m.id in used_modules or m.module.approval_status == PvcPVModule.PV_MODULE_STATUS_REJECTED:
                    continue
                if m.subrogation is not None and m.commissioner_approval_status == PvcTbsMemberPVModule.COMMISSIONER_APPROVAL_STATUS_REJECTED:
                    continue
                possible['m_%s' % m.id] = m

            additional = [value for key, value in additional_lines.iteritems() if key[:18]=='tbs_member_module_']
            # each line should be unique on tbs_member_module level.
            for nl in additional:

                if nl in used_modules:
                    raise PublicUserException(_('New line contains a module which is already part of this Declaration.'))
                if not possible.has_key('m_%s' % nl):
                    raise PublicUserException(_('New line contains a module which is not active for this Declaration.'))

                module = possible['m_%s' % nl]
                quantity = int(additional_lines['quantity_additional_%s' % nl])
                if quantity < 0:
                    raise PublicUserException(_('All quantities must be larger or equal to zero.'))

                lines.append({
                    'polymorphic_type': PvcDeclarationOrderLine,
                    'description': '%s - %s' % (module.module.supplier.name, module.module.type.name),
                    'quantity': quantity,
                    'price': af,
                    'tbs_member_module': module,
                })

        order = self._save_or_update_order(order.id, lines, order.merchant, order.client.default_group)
        order.status = self._instance_class.STATUS_PENDING
        order.is_compliant = self._instance_class.IS_COMPLIANT # if the declaration got updated, they are compliant

        self.sa_session.add(order)
        return order

    @transaction_decorator
    def remove_line(self, _id, line_id, disable_module=None):
        if isinstance(_id, self._instance_class):
            order = _id
        else:
            orders = self.get(_id, priv=self._instance_class.PRIV_UPDATE)
            if len(orders) < 1:
                raise PublicSecurityException(_('Cannot update Order #%s' % _id))
            order = orders.pop()

        lines = [l for l in order.lines if l.id == line_id]
        if len(lines) == 0:
            raise PublicUserException(_('Cannot find line within order.'))

        line = lines[0]
        self._assert_valid_line_removal(order, line)
        if line.tbs_member_module.module.approval_status == PvcPVModule.PV_MODULE_STATUS_REJECTED:
            disable_module = PvcTbsMemberPVModule.IS_OBSOLETE

        if disable_module is not None:
            c = PvcPVModuleController(self.sa_session, account_id=self.account_id, request=self.request)
            c.change_tbs_member_module_active_state(line.tbs_member_module, disable_module)

        return super(PvcDeclarationOrderController, self).remove_line(order, line_id)

    @transaction_decorator
    def resubmit_declaration(self, order, quantities, additional_lines=None):
        assert isinstance(order, self._instance_class)
        if not self.has_different_quantities(order, quantities) and additional_lines is None:
            # make sure the order is set to compliant to prevent issue with minimum contribution fee
            order.is_compliant = PvcDeclarationOrder.IS_COMPLIANT
            self.sa_session.add(order)
            return order

        order.resubmit_request = None
        self.sa_session.add(order)

        credit = self._credit_order(order)
        order = self._create_child_order(credit, exclude_line_type=PvcDeclarationOrderLineMinimumCF)
        return self.update_declaration(order.id, quantities, additional_lines)

    def _assert_ready_for_approval(self, order):
        """
        If one or more PV Modules are not approved, we block this Declaration. See ticket #152.
        """
        for line in order.lines:
            if isinstance(line, PvcDeclarationOrderLineMinimumCF):
                continue
            module = line.tbs_member_module.module
            if line.tbs_member_module.subrogation is not None and line.tbs_member_module.commissioner_approval_status != PvcTbsMemberPVModule.COMMISSIONER_APPROVAL_STATUS_APPROVED:
                raise PvcPVModuleNotApprovedException(_('The PV Module "%(supplier)s - %(type)s" for member "%(member)s" at TBS "%(tbs)s" is not approved by the Commissioner.' % {
                    'supplier': module.supplier.name,
                    'type': module.type.name,
                    'member': order.client.name,
                    'tbs': order.merchant.name
                }))
            if module.approval_status != PvcPVModule.PV_MODULE_STATUS_APPROVED:
                raise PvcPVModuleNotApprovedException(_('The PV Module "%(supplier)s - %(type)s" for member "%(member)s" at TBS "%(tbs)s" is not approved.' % {
                    'supplier': module.supplier.name,
                    'type': module.type.name,
                    'member': order.client.name,
                    'tbs': order.merchant.name
                }))

    def _get_price(self, tbs_member):
        asl = tbs_member.active_service_level
        if asl is None:
            raise PublicUserException(_('Cannot add declarations for a Member without an active Service Level.'))
        af = asl.service_level.get_current_contribution_fee()
        if af is None:
            raise PublicUserException(_('Cannot add declarations for a Member without a Contribution Fee for this Service Level'))
        return af

    def get_new_lines(self, _id, module_id):
        self.assert_permissions('pvc_declare_declarations')
        order = self.get(_id, priv=self._instance_class.PRIV_UPDATE).pop()
        modules = order.producer.modules

        declaration_modules = [l.tbs_member_module for l in order.lines if hasattr(l, 'tbs_member_module')]

        _new = [module for module in modules if module not in declaration_modules]
        # dealing with subrogation
        if order.producer != order.tbs_member:
            # filtering only for the correct commmissioner
            _new = [module for module in _new if module.subrogation is not None and module.subrogation.commissioner == order.tbs_member]
        else:
            # filtering only for own producer modules (subrogation should be None)
            _new = [module for module in _new if module.subrogation is None]

        if len(_new) == 0:
            return []

        rv = []
        for tbs_member_module in _new:

            module = tbs_member_module.module
            if module.approval_status == PvcPVModule.PV_MODULE_STATUS_REJECTED:
                continue
            if tbs_member_module.subrogation is not None and tbs_member_module.commissioner_approval_status == PvcTbsMemberPVModule.COMMISSIONER_APPROVAL_STATUS_REJECTED:
                continue
            price = self._get_price(order.producer)
            tax_rate, vat_fee = self._get_vat_rate(False, self._instance_class, order.merchant, price)

            assert tax_rate is not None
            assert vat_fee is not None

            rv.append({
                'tbs_member_module_id': tbs_member_module.id,
                'description': '%s - %s' % (module.supplier.name, module.type.name),
                'price_per_unit': price,
                'quantity': 0,
                'vat_rate': tax_rate,
                'gl_account': None,
                'line_amount_excluding_vat': 0,
                'line_amount_including_vat': 0,
                'supplier': module.supplier.name,
                'type': module.type.name,
                'cell_type': module.cell_type.name,
                'nominal_power': module.nominal_power,
                'weight': module.weight,
                'country': tbs_member_module.country.name,
                'surface': module.width*module.length,
                'module_status': module.approval_status,
            })
        return rv

    def _is_undeclared_period(self, tbs_member, year, quarter, month, producer):

        order = self.sa_session.query(PvcDeclarationOrder).filter(
            and_(
                PvcDeclarationOrder.tbs_member==tbs_member,
                PvcDeclarationOrder.year==year,
                PvcDeclarationOrder.quarter==quarter,
                PvcDeclarationOrder.month==month,
                PvcDeclarationOrder.producer==producer
            )
        ).first()

        if order is None:
            return True

        return False

    def get_available_subrogations_for_order(self, _id):
        """
        This returns a list with Subrogations which do not yet have a Declaration for the same period as the supplied order.
        """
        orders = self.get(_id, priv=self._instance_class.PRIV_UPDATE)
        if len(orders) == 0: # this happens when the load state requester is the COMMISSIONER in case of subrogation
            return []
        order = orders.pop()

        if order.producer != order.tbs_member:
            return []

        if not self.has_active_subrogations(order.tbs_member):
            return []

        rv = []
        for subrogation in order.tbs_member.subrogations:
            if subrogation.status != PvcSubrogation.STATUS_ACTIVE:
                continue

            if not self._is_undeclared_period(subrogation.commissioner, order.year, order.quarter, order.month, order.producer):
                continue
            rv.append(subrogation)

        return rv

    def get_undeclared_periods(self, tbs_member):
        lcc = LicenseConfigController(self.sa_session, scope=tbs_member.tbs)
        period = lcc.get('pvc.tbs.declaration.period')
        if period == PvcTbsLicense.DECLARATION_PERIOD_YEARLY:
            min_year = tbs_member.created.year
            possible = self._get_possible_periods_for_yearly_declarations(min_year)
            if possible is None:
                return {}

            declarations = self.sa_session.query(PvcDeclarationOrder).filter(PvcDeclarationOrder.tbs_member==tbs_member).all()
            for d in declarations:
                if not possible.has_key(d.year):
                    continue

                del possible[d.year]
            return possible

        if period == PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY:
            min_year = tbs_member.created.year
            possible = self._get_possible_periods_for_quarterly_declarations(min_year)
            if possible is None:
                return {}

            declarations = self.sa_session.query(PvcDeclarationOrder).filter(PvcDeclarationOrder.tbs_member==tbs_member).all()
            for d in declarations:
                if not possible.has_key(d.year):
                    continue
                if d.quarter not in possible[d.year]['quarters']:
                    continue
                possible[d.year]['quarters'] = filter(lambda a: a != d.quarter, possible[d.year]['quarters'])
            rm_y = []
            for year, quarters in possible.iteritems():
                if len(quarters['quarters']) == 0:
                    rm_y.append(year)
            for year in rm_y:
                del possible[year]
            return possible

        if period == PvcTbsLicense.DECLARATION_PERIOD_MONTHLY:

            min_year = tbs_member.created.year
            possible = self._get_possible_periods_for_monthly_declarations(min_year)
            if possible is None:
                return {}

            declarations = self.sa_session.query(PvcDeclarationOrder).filter(PvcDeclarationOrder.tbs_member==tbs_member).all()
            for d in declarations:
                if not possible.has_key(d.year):
                    continue
                if d.month not in possible[d.year]['months']:
                    continue
                possible[d.year]['months'] = filter(lambda a: a != d.month, possible[d.year]['months'])
            rm_y = []
            for year, months in possible.iteritems():
                if len(months['months']) == 0:
                    rm_y.append(year)
            for year in rm_y:
                del possible[year]
            return possible

        raise SecurityException('Incorrect period set for TBS: %s' % tbs_member.tbs)

    def _get_possible_periods_for_yearly_declarations(self, min_year):
        td = today()
        max_year = td.year-1
        if td.month == 1 and td.day == 1:
            max_year = max_year-1
        if max_year < min_year:
            return None

        years = range(min_year, max_year+1)
        years.reverse()

        rv = {}
        for year in years:
            rv[year] = {}

        return rv

    def _get_possible_periods_for_quarterly_declarations(self, min_year):
        td = today()
        max_year = td.year
        import math
        max_q = int(math.floor((td.month-1)/3))
        if td.month%3 == 1 and td.day == 1:
            max_q = max_q-1
        if max_q < 1:
            max_q = 4
            max_year = max_year-1
        if max_year < min_year:
            return None

        years = range(min_year, max_year+1)
        years.reverse()

        rv = {}
        for year in years:
            quarters = []
            while max_q > 0:
                quarters.insert(0, max_q)
                max_q -= 1
            rv[year] = {'quarters': quarters}
            max_q = 4

        return rv

    def _get_possible_periods_for_monthly_declarations(self, min_year):
        td = today()
        max_year = td.year
        max_month = td.month-1 if td.day > 1 else td.month-2
        if max_month < 1:
            max_month = 12
            max_year = max_year-1
        if max_year < min_year:
            return None

        years = range(min_year, max_year+1)
        years.reverse()

        rv = {}
        for year in years:
            months = []
            while max_month > 0:
                months.insert(0, max_month)
                max_month -= 1
            rv[year] = {'months': months}
            max_month = 12

        return rv

    @transaction_decorator
    def add_open_declaration(self, tbs_member, year, month=None, quarter=None, producer=None):
        """
        This will add a new Declaration for the provided TBS Member in the year supplied.
        Checks on month first quarter depending on the TBS setting.
        Raises an exception in case there already is a Declaration.

        Credits/Resubmits are done through other methods.
        """

        self.logger.info('add_open_declaration: tbs_member: %s, year=%s, month=%s, quarter=%s' % (tbs_member, year, month, quarter))

        assert year is not None
        if month == '':
            month = None
        if quarter == '':
            quarter = None

        lcc = LicenseConfigController(self.sa_session, scope=tbs_member.tbs)
        period = lcc.get('pvc.tbs.declaration.period')

        if period == PvcTbsLicense.DECLARATION_PERIOD_YEARLY:
            assert month is None and quarter is None
        elif period == PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY:
            assert month is None and quarter is not None
        elif period == PvcTbsLicense.DECLARATION_PERIOD_MONTHLY:
            assert quarter is None and month is not None
        else:
            raise SecurityException('Incorrect period set for TBS: %s' % tbs_member.tbs)

        # TODO: can only add for current year and previous periods OR previous year if period is YEARLY
        if producer is None:
            producer = tbs_member

        order = self.sa_session.query(PvcDeclarationOrder).filter(and_(
            PvcDeclarationOrder.tbs_member==tbs_member,
            PvcDeclarationOrder.year==year,
            PvcDeclarationOrder.quarter==quarter,
            PvcDeclarationOrder.month==month,
            PvcDeclarationOrder.producer==producer,
        )).first()
        if order is not None:
            raise PublicSecurityException(_('Declaration already exists for provided TBS Member and period.'))

        af = self._get_price(tbs_member)

        orders = {}

        lines = []
        for module in tbs_member.modules:
            if module.is_active != PvcTbsMemberPVModule.IS_ACTIVE: # we only add active modules
                continue

            s = module.subrogation
            if s is not None and module.commissioner_approval_status != PvcTbsMemberPVModule.COMMISSIONER_APPROVAL_STATUS_APPROVED:
                continue

            if not orders.has_key(s):
                orders[s] = {
                    'merchant': tbs_member.tbs,
                    'scope': tbs_member.member.default_group if s is None else module.subrogation.commissioner.member.default_group,
                    'tbs_member': tbs_member if s is None else module.subrogation.commissioner,
                    'year': year,
                    'month': month,
                    'quarter': quarter,
                    'producer': tbs_member,
                    'lines': []
                }

            orders[s]['lines'].append({
                'polymorphic_type': PvcDeclarationOrderLine,
                'description': '%s - %s' % (module.module.supplier.name, module.module.type.name),
                'quantity': 0,
                'price': af,
                'tbs_member_module': module,
            })

        rv = []
        for s, order in orders.iteritems():
            if len(order['lines']) == 0:
                self.logger.warn('add_open_declaration: NO ACTIVE PV MODULES for tbs_member: %s, year=%s, month=%s, quarter=%s, subrogation=%s' % (tbs_member, year, month, quarter, s))
                continue

            self.logger.info('add_open_declaration: DECLARATION ADDED for tbs_member: %s, year=%s, month=%s, quarter=%s, subrogation=%s' % (tbs_member, year, month, quarter, s))
            rv.append(self._save_or_update_order(None, **order))

        if len(rv) == 0:
            return None

        if len(rv) == 1:
            return rv[0]

        return rv

    def _get_line_unit(self, quantity, order, **kwargs):

        if self.merchant_uses_weight is None:
            self.merchant_uses_weight = {}
        if not self.merchant_uses_weight.has_key(order.merchant.id):
            lcc = LicenseConfigController(self.sa_session, scope=order.merchant)
            self.merchant_uses_weight[order.merchant.id] = lcc.get('pvc.tbs.declaration.use.weight.for.amount.calculations')
        if self.merchant_uses_weight[order.merchant.id] and kwargs.has_key('tbs_member_module'):
            # if based on weight, the weight should be multiplied by the number of units
            return kwargs['tbs_member_module'].module.weight*quantity

        return quantity

    def get_next_closing_date(self, tbs):

        today = virtual_date.today()

        lcc = LicenseConfigController(self.sa_session, scope=tbs)
        period = lcc.get('pvc.tbs.declaration.period')
        closing_day = lcc.get('pvc.tbs.declaration.closing.day')
        if closing_day is None:
            raise Exception('Closing day is None')

        closing_date = None

        if period == PvcTbsLicense.DECLARATION_PERIOD_YEARLY:
            closing_date = datetime.date(today.year, 1, 1) + datetime.timedelta(days=closing_day)

            if closing_date < today:
                closing_date = closing_date.replace(year=today.year+1)

        elif period == PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY:

            # get current quarter
            q = range(4)
            q.reverse()
            for i in q:
                x = datetime.date(today.year, (i*3)+1, 1)
                if x <= today:
                    break
            closing_date = datetime.date(x.year, x.month, 1) + datetime.timedelta(days=closing_day)

            if closing_date < today:
                closing_date = closing_date + relativedelta(months=3)

        elif period == PvcTbsLicense.DECLARATION_PERIOD_MONTHLY:
            closing_date = datetime.date(today.year, today.month, 1) + datetime.timedelta(days=closing_day)
            if closing_date < today:
                closing_date = closing_date + relativedelta(months=1)

        assert closing_date != None
        return closing_date

    def _close_declarations(self, open_orders):
        closed = []
        for order in open_orders:
            try:
                self.submit_order(order)
                closed.append(order)
                self.logger.info('_close_declarations: closing declaration %s' % order)
            except PvcPVModuleNotApprovedException as e:
                self.logger.info(u"_close_declarations: not closing declaration %s, unapproved module: %s" % (order, unicode(e)))
                nc = PvcNotificationController(self.sa_session)
                note = nc.set_notification(_('Unapproved PV Module blocked Declaration from closing.'), unicode(e))
                groups = self.get_groups_in_domain_with_permissions('pvc_approve_pv_modules', domain_id=order.merchant.domain.id, anonymous=True)
                if len(groups) == 0:
                    self.logger.info(u"_close_declarations: unable to send system notification for order %s" % order)
                    continue
                nc.create_system_notification(note, groups, for_group=True)
        return closed

    @transaction_decorator
    def close_declarations(self, tbs):

        today = virtual_date.today()

        if (today != self.get_next_closing_date(tbs)):
            self.logger.debug('today is not closing day for tbs %s' % tbs)
            return None

        _where = self._instance_class.status.in_([self._instance_class.STATUS_NEW, self._instance_class.STATUS_PENDING])
        _where = and_(_where, self._instance_class.type!=self._instance_class.TYPE_CREDIT)
        _where = and_(_where, self._instance_class.merchant==tbs)

        open_orders = self.sa_session.query(self._instance_class).filter(_where).all()
        closed = self._close_declarations(open_orders)
        return closed

    @transaction_decorator
    def close_declarations_created_before_specified_date(self, tbs, before):
        self.assert_permissions('pvc_approve_declaration_resubmit', 'pvc_tbs_license')
        today = virtual_date.today()
        max_dt = datetime.date(today.year, today.month, today.day)
        if before.date() >= max_dt:
            raise PublicUserException('Invalid period for closing')

        _where = self._instance_class.status.in_([self._instance_class.STATUS_NEW, self._instance_class.STATUS_PENDING])
        _where = and_(_where, self._instance_class.type!=self._instance_class.TYPE_CREDIT)
        _where = and_(_where, self._instance_class.merchant==tbs)
        _where = and_(_where, self._instance_class.created<before)

        self.logger.info('manual trigger for closing all declarations for TBS %s before %s' % (tbs, before))

        open_orders = self.sa_session.query(self._instance_class).filter(_where).all()
        closed = self._close_declarations(open_orders)
        return closed

    @transaction_decorator
    def update_with_new_fee(self, service_level, new_fee, old_fee):
        # complexity:
        #  * There is no direct relationship between service levels and declarations
        #  * We will only update Declarations for Producers who have the provided Service Level set as their current SL.

        tbs = service_level.tbs
        tbs_members = self.sa_session.query(PvcTbsMember).filter(and_(PvcTbsMember.tbs==tbs, PvcTbsMember.is_producer==True)).all()
        _in = []
        for tm in tbs_members:
            if tm.active_service_level.service_level==service_level:
                _in.append(tm.id)

        # get all declarations for this TBS where the TBS Members have the selected service level
        _where = self._instance_class.status==self._instance_class.STATUS_NEW
        _where = and_(_where, self._instance_class.type!=self._instance_class.TYPE_CREDIT)
        _where = and_(_where, self._instance_class.merchant==tbs)
        _where = and_(_where, self._instance_class.tbs_member_id.in_(_in))

        orders = self.sa_session.query(self._instance_class).filter(_where).all()
        affected = []
        for order in orders:
            for line in order.lines:
                if line.price_per_unit != old_fee:
                    raise PublicUserException(_('Incorrect original fee found for order #%(order)d on line %(line)d' % {'order': order.id, 'line': line.id}))
                if line.quantity != 0:
                    raise PublicUserException(_('Found quantities on order #%(order)d on line %(line)d with status new' % {'order': order.id, 'line': line.id}))
                line.price_per_unit = new_fee
                self.sa_session.add(line)
                self.logger.info('altering price per unit for order #%d on line %s from fee %s to %s' % (order.id, line.id, str(old_fee), str(new_fee)))
            affected.append(order)
        return affected

    def trigger_new_declarations(self, tbs):
        """
        triggers new declarations for the given tbs by checking the declaration period
        """

        today = virtual_date.today()
        if today.day != 1: # we trigger only on the first
            return None

        lcc = LicenseConfigController(self.sa_session, scope=tbs)
        period = lcc.get('pvc.tbs.declaration.period')

        year = None
        month = None
        quarter = None

        if period == PvcTbsLicense.DECLARATION_PERIOD_YEARLY:
            if not (today.month == 1 and today.day == 1):
                return None
            year = today.year-1

        elif period == PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY:

            if not (today.month in (1, 4, 7, 10) and today.day == 1):
                return None

            # get previous quarter
            q = None
            r = range(4)
            r.reverse()
            for i in r:
                x = datetime.date(today.year, (i*3)+1, 1)
                if x < today:
                    q = x
                    year = today.year
                    quarter = i+1
                    break
            if q is None:
                q = datetime.date(today.year-1, 10, 1)
                year = today.year-1
                quarter = 4

        elif period == PvcTbsLicense.DECLARATION_PERIOD_MONTHLY:
            month = today.month -1
            year = today.year

            if month == 0:
                month = 12
                year = today.year -1

        else:
            raise SecurityException('Incorrect period set for TBS: %s' % tbs)

        tbs_members = self.sa_session.query(PvcTbsMember)\
            .filter(PvcTbsMember.tbs == tbs,
                    PvcTbsMember.is_producer == True,
                    PvcTbsMember.status.in_( (PvcTbsMember.STATUS_APPROVED, PvcTbsMember.STATUS_RESIGNING, ) )).all()

        orders = []
        for tbs_member in tbs_members:

            try:
                order = self.add_open_declaration(tbs_member, year, month, quarter)
                if order is not None:
                    orders.append(order)

            # This is trigger from the scheduler and we do not want public security exceptions to block the process
            # catch public user exceptions for 'Cannot add declarations for a Member without an active Service Level
            except (PublicUserException, PublicSecurityException), e:
                self.logger.warn('add_open_declaration: %s' % e.message)

        return orders

    @transaction_decorator
    def submit_order(self, _id):
        """
        extends super.submit_order() and sets the display_date
        ticket http://trac.office.basesoft.com/iqportal/ticket/3974
        """
        order = super(PvcDeclarationOrderController, self).submit_order(_id)
        order.display_date = now()

        if order.is_compliant is None:
            compliant = PvcDeclarationOrder.IS_NON_COMPLIANT
            if self.get_account() is not None or order.total_amount_excluding_vat > 0:
                compliant = PvcDeclarationOrder.IS_COMPLIANT

            order.is_compliant = compliant
        self.sa_session.add(order)
        order = self.add_minimum_contribution_fee(order)
        return order

    @transaction_decorator
    def add_minimum_contribution_fee(self, order):
        #get original order and if order is compliant we don't need to recalculate the min CF
        if self._get_original_order(order).is_compliant:
            return order

        lcc = LicenseConfigController(self.sa_session, scope=order.merchant)
        minimum_fee = lcc.get('pvc.tbs.minimum.contribution.fee')
        if minimum_fee == 0:
            return order
        #if declared amount is bigger than minimum Contribution fee forfeit fee is canceled
        #if declared amount is smaller a new line is created
        #with the difference between the min Contribution fee and declared amount
        if order.total_amount_excluding_vat > minimum_fee:
            return order
        minimum_fee = minimum_fee - order.total_amount_excluding_vat

        country = order.merchant.company.addresses[0].address.country
        client_country = order.client.company.addresses[0].address.country

        apply_vat = client_country == country
        self._save_or_update_order_line(order, len(order.lines), order.merchant, 'Minimum contribution fee', quantity=1, price=minimum_fee, polymorphic_type=PvcDeclarationOrderLineMinimumCF, apply_vat=apply_vat)

        order.total_amount_excluding_vat = sum([l.line_amount_excluding_vat for l in order.lines])
        order.total_amount_including_vat = sum([l.line_amount_including_vat for l in order.lines])
        order.invoiced_amount = self._get_invoiced_amount(order)
        self.sa_session.add(order)
        return order

    def send_declaration_reminder_email(self, order_ids):

        self.assert_permissions('pvc_manage_declaration_mail')

        from basesoft.web.contrib.mailing.schedule import HtmlMailTask
        mail_task = HtmlMailTask()

        for order_id in order_ids:

            declaration = self.get_order(order_id)[0]
            mail = self.get_object_by_priv_in_hierarchy(declaration.client.default_group, PvcDeclarationReminderMail, PvcDeclarationReminderMail.PRIV_USE_FOR_SENDING)
            if mail is None:
                raise PublicUserException(_('No declaration reminder mail is available for sending.'))

            license_ = declaration.client
            person = declaration.client.company.contact_person

            msg, result = mail_task.build_message(person, mail, context={
                'license': license_,
                'declaration': declaration
            }, mail_from=mail.mail_from())
            mail_task.send_mail(person, msg, result, mail.mail_from())

    def __get_period_label(self, order):
        period = order.year
        if order.month is not None:
            period = '%s-%s' % (order.month, order.year)
        elif order.quarter is not None:
            period = 'Q%s-%s' % (order.quarter, order.year)
        return period

    @transaction_decorator
    def request_for_resubmission(self, _id):
        self.assert_permissions('pvc_declare_declarations')
        order = self.get(_id, priv=self._instance_class.PRIV_VIEW).pop()
        if not self.has_privileges_on_object(order, self._instance_class.PRIV_UPDATE):
            if order.tbs_member != order.producer:
                raise PublicUserException(_('A resubmit request can only be done by the producer.'))
            raise PublicUserException(_('You cannot make a resubmit request for this order.'))

        if order.status != self._instance_class.STATUS_SUBMITTED:
            raise PublicUserException(_('You cannot make a resubmit request for an unsubmitted order.'))

        if order.resubmit_request is not None:
            raise PublicUserException(_('No resubmit request possible for this order.'))

        order.resubmit_request = self._instance_class.RESUBMIT_REQUEST_PENDING
        self.sa_session.add(order)

        period = self.__get_period_label(order)

        nc = PvcNotificationController(self.sa_session, account_id=self.get_account().id)
        operators = self.sa_session.query(PvcOperatorLicense).filter(PvcOperatorLicense.domain==order.tbs_member.tbs.domain).all()
        recipients = self.get_groups_in_license_with_permissions('pvc_tbs_member_manager', license_id=order.tbs_member.tbs_id)
        recipients = recipients + [o.default_group for o in operators]
        message = 'The member "%(member)s" at TBS "%(tbs)s" has made a request for resubmission for period "%(period)s".' % {'member': order.tbs_member.member.name, 'tbs': order.tbs_member.tbs.name, 'period': period}
        notification = nc.set_notification('Resubmit request by member "%(member)s" at TBS "%(tbs)s".' % {'member': order.tbs_member.member.name, 'tbs': order.tbs_member.tbs.name}, message)
        nc.create_system_notification(notification, recipients, for_license=True)
        return True

    @transaction_decorator
    def approve_resubmit_request(self, _id):
        self.assert_permissions('pvc_approve_declaration_resubmit')
        order = self.get(_id, priv=self._instance_class.PRIV_VIEW).pop()

        if order.resubmit_request != self._instance_class.RESUBMIT_REQUEST_PENDING:
            raise PublicUserException(_('This order (#%s) does not have an outstanding resubmit request.' % order.id))

        order.resubmit_request = self._instance_class.RESUBMIT_REQUEST_APPROVED
        self.sa_session.add(order)

        period = self.__get_period_label(order)

        nc = PvcNotificationController(self.sa_session, account_id=self.get_account().id)
        recipients = self.get_groups_in_license_with_permissions('pvc_member_license', license_id=order.tbs_member.member_id)
        notification = nc.set_notification('Resubmit request approved', 'Your request for resubmission at TBS "%(tbs)s" for period "%(period)s" has been approved. Please navigate to the declaration overview to make your resubmission.' % {'tbs': order.tbs_member.tbs.name, 'period': period})
        nc.create_system_notification(notification, recipients, for_license=True)

        return True

    @transaction_decorator
    def reject_resubmit_request(self, _id, reason):
        self.assert_permissions('pvc_approve_declaration_resubmit')
        order = self.get(_id, priv=self._instance_class.PRIV_VIEW).pop()

        if order.resubmit_request != self._instance_class.RESUBMIT_REQUEST_PENDING:
            raise PublicUserException(_('This order (#%s) does not have an outstanding resubmit request.' % order.id))

        order.resubmit_request = None
        self.sa_session.add(order)

        period = self.__get_period_label(order)

        nc = PvcNotificationController(self.sa_session, account_id=self.get_account().id)
        recipients = self.get_groups_in_license_with_permissions('pvc_member_license', license_id=order.tbs_member.member_id)
        notification = nc.set_notification('Resubmit request rejected', 'Your request for resubmission at TBS "%(tbs)s" for period "%(period)s" has been rejected with the following reason: %(reason)s.' % {
            'tbs': order.tbs_member.tbs.name,
            'period': period,
            'reason': reason
        })
        nc.create_system_notification(notification, recipients, for_license=True)

        return True

class PvcDeclarationAggregator(IQubeAggregator):
    _table = "aggPvcDeclaration"
    _affected_cubes = ({'datasource': 'PvcOrderData', 'cube': 'PvcDeclarationDataCube'}, )
    _columns = [
        'order_id',
        'order_line_id',
        'order_type',
        'order_status',
        'invoice_type',
        'country_id',
        'year',
        'quarter',
        'month',
        'tbs_id',
        'member_id',
        'producer_id',
        'pv_module_id',
        'supplier_id',
        'submitted_time_id',
        'payment_status',
        'payment_method',
        'payment_time_id',
        'gl_account',
        'quantity',
        'weight',
        'power',
        'line_amount_excluding_vat',
        'line_amount_including_vat',
        'price_per_unit',
        'last_updated'
    ]

    _select_sql = """

        SELECT
            o.id as order_id,
            ol.id as order_line_id,
            o.polymorphic_type as order_type,
            o.status as order_status,
            IF(o.type=0, 'Debit', 'Credit') as invoice_type,
            mpvm.country_id as country_id,
            do.year as `year`,
            IF(ISNULL(do.quarter) AND NOT ISNULL(do.month), IF(do.month < 4, 1, IF(do.month < 7, 2, IF(do.month < 10, 3, 4))), do.quarter) as `quarter`,
            do.month as `month`,
            ptm.member_id as producer_id,
            o.client_id as member_id,
            o.merchant_id as tbs_id,
            pvm.id as pv_module_id,
            pvms.id as supplier_id,
            IF(ISNULL(o.display_date), NULL, DATE_FORMAT(o.display_date, '%%Y%%m%%d')) as submitted_time_id,
            iqGetCurrentOrderPaymentStatus(o.id) as payment_status,
            iqGetCurrentOrderPaymentMethod(o.id) as payment_method,
            IF(ISNULL(iqGetCurrentOrderPaymentDate(o.id)), NULL, DATE_FORMAT(o.display_date, '%%Y%%m%%d')) as payment_time_id,
            ol.gl_account_number as gl_account,
            ol.quantity as quantity,
            (ol.quantity * pvm.weight) as weight,
            (ol.quantity * pvm.nominal_power) as power,
            ol.line_amount_excluding_vat as line_amount_excluding_vat,
            ol.line_amount_including_vat as line_amount_including_vat,
            ol.price_per_unit as price_per_unit,
            o.last_updated as last_updated
        FROM tblOrder as o
        INNER JOIN tblOrderLine as ol
            ON o.id = ol.order_id
        INNER JOIN tblPvcDeclarationOrder as do
            ON o.id = do.order_id
        INNER JOIN tblPvcDeclarationOrderLine as dol
            ON ol.id = dol.order_line_id
        INNER JOIN tblPvcTbsMember as ptm
            ON do.producer_id = ptm.id
        INNER JOIN tblPvcTbsMemberPVModule as mpvm
            ON dol.tbs_member_module_id = mpvm.id
        INNER JOIN tblPvcPVModule as pvm
            ON mpvm.module_id = pvm.id
        INNER JOIN tblPvcPVModuleSupplier as pvms
            ON pvm.supplier_id = pvms.id
        WHERE o.last_updated > %(max_last_updated)s
        AND o.deleted = 0
        ORDER BY o.id
    """

    def run(self, loop_validator=None):

        affected_rows = 0

        max_last_updated = 0
        rows = self.db_engine_ro.execute("""
            SELECT IFNULL(MAX(a.last_updated), 0)
            FROM %s as a""" % self._table)

        row = rows.fetchone()
        rows.close()

        if row is not None:
            max_last_updated = row[0]

        self.db_engine.execute("""
            DELETE FROM %(tbl)s WHERE order_id IN (
                SELECT id
                FROM tblOrder
                WHERE last_updated > '%(max_last_updated)s'
            )""" % {
            'tbl': self._table,
            'max_last_updated': max_last_updated
        })

        affected_rows = affected_rows + self._work({
            'max_last_updated': max_last_updated,
        }, loop_validator)

        return affected_rows

